import math
from typing import List, Sequence, Tuple, Union

import torch
import torch.nn as nn
from mmdet.models.utils import multi_apply
from mmdet.utils import (ConfigType,
                         OptMultiConfig)
from mmengine.model import BaseModule
from torch import Tensor
from mmengine.model import BaseModule
from ..layers.yolo_bricks import DWConv, Conv
from mmengine.registry import MODELS
from mmyolo.models.utils import make_divisible, make_round

@MODELS.register_module()
class YOLO11HeadModule(BaseModule):
    """YOLOv8HeadModule head module used in `YOLOv8`.

    Args:
        num_classes (int): Number of categories excluding the background
            category.
        in_channels (Union[int, Sequence]): Number of channels in the input
            feature map.
        widen_factor (float): Width multiplier, multiply number of
            channels in each layer by this amount. Defaults to 1.0.
        num_base_priors (int): The number of priors (points) at a point
            on the feature grid.
        featmap_strides (Sequence[int]): Downsample factor of each feature map.
             Defaults to [8, 16, 32].
        reg_max (int): Max value of integral set :math: ``{0, ..., reg_max-1}``
            in QFL setting. Defaults to 16.
        norm_cfg (:obj:`ConfigDict` or dict): Config dict for normalization
            layer. Defaults to dict(type='BN', momentum=0.03, eps=0.001).
        act_cfg (:obj:`ConfigDict` or dict): Config dict for activation layer.
            Defaults to None.
        init_cfg (:obj:`ConfigDict` or list[:obj:`ConfigDict`] or dict or
            list[dict], optional): Initialization config dict.
            Defaults to None.
    """

    def __init__(self,
                 num_classes: int,
                 in_channels: Union[int, Sequence],
                 widen_factor: float = 1.0,
                 num_base_priors: int = 1,
                 featmap_strides: Sequence[int] = (8, 16, 32),
                 reg_max: int = 16,
                 init_cfg: OptMultiConfig = None):
        super().__init__(init_cfg=init_cfg)
        self.num_classes = num_classes
        self.featmap_strides = featmap_strides
        self.num_levels = len(self.featmap_strides)
        self.num_base_priors = num_base_priors
        self.in_channels = in_channels
        self.reg_max = reg_max

        in_channels = []
        for channel in self.in_channels:
            channel = make_divisible(channel, widen_factor)
            in_channels.append(channel)
        self.in_channels = in_channels

        self._init_layers()

    def init_weights(self, prior_prob=0.01):
        """Initialize the weight and bias of PPYOLOE head."""
        super().init_weights()
        for reg_pred, cls_pred, stride in zip(self.reg_preds, self.cls_preds,
                                              self.featmap_strides):
            reg_pred[-1].bias.data[:] = 1.0  # box
            # cls (.01 objects, 80 classes, 640 img)
            cls_pred[-1].bias.data[:self.num_classes] = math.log(
                5 / self.num_classes / (640 / stride)**2)

    def _init_layers(self):
        """initialize conv layers in YOLOv8 head."""
        # Init decouple head
        self.cls_preds = nn.ModuleList()
        self.reg_preds = nn.ModuleList()

        reg_out_channels = max(
            (16, self.in_channels[0] // 4, self.reg_max * 4))
        cls_out_channels = max(self.in_channels[0], self.num_classes)

        for i in range(self.num_levels):
            self.reg_preds.append(
                nn.Sequential(
                    Conv(
                        c1=self.in_channels[i],
                        c2=reg_out_channels,
                        k=3),
                    Conv(
                        c1=reg_out_channels,
                        c2=reg_out_channels,
                        k=3,),
                    nn.Conv2d(
                        in_channels=reg_out_channels,
                        out_channels=4 * self.reg_max,
                        kernel_size=1)))
            self.cls_preds.append(
                nn.Sequential(
                    nn.Sequential(
                        DWConv(
                            c1=self.in_channels[i],
                            c2=self.in_channels[i],
                            k=3),
                        Conv(
                            c1=self.in_channels[i],
                            c2=cls_out_channels,
                            k=1
                        ),),
                    nn.Sequential(
                        DWConv(
                            c1=cls_out_channels,
                            c2=cls_out_channels,
                            k=3),
                        Conv(
                            c1=cls_out_channels,
                            c2=cls_out_channels,
                            k=1
                        ),),
                    nn.Conv2d(
                        in_channels=cls_out_channels,
                        out_channels=self.num_classes,
                        kernel_size=1)))

        proj = torch.arange(self.reg_max, dtype=torch.float)
        self.register_buffer('proj', proj, persistent=False)

    def forward(self, x: Tuple[Tensor]) -> Tuple[List]:
        """Forward features from the upstream network.

        Args:
            x (Tuple[Tensor]): Features from the upstream network, each is
                a 4D-tensor.
        Returns:
            Tuple[List]: A tuple of multi-level classification scores, bbox
            predictions
        """
        assert len(x) == self.num_levels
        return multi_apply(self.forward_single, x, self.cls_preds,
                           self.reg_preds)

    def forward_single(self, x: torch.Tensor, cls_pred: nn.ModuleList,
                       reg_pred: nn.ModuleList) -> Tuple:
        """Forward feature of a single scale level."""
        b, _, h, w = x.shape
        cls_logit = cls_pred(x)
        bbox_dist_preds = reg_pred(x)
        if self.reg_max > 1:
            bbox_dist_preds = bbox_dist_preds.reshape(
                [-1, 4, self.reg_max, h * w]).permute(0, 3, 1, 2)

            # TODO: The get_flops script cannot handle the situation of
            #  matmul, and needs to be fixed later
            # bbox_preds = bbox_dist_preds.softmax(3).matmul(self.proj)
            bbox_preds = bbox_dist_preds.softmax(3).matmul(
                self.proj.view([-1, 1])).squeeze(-1)
            bbox_preds = bbox_preds.transpose(1, 2).reshape(b, -1, h, w)
        else:
            bbox_preds = bbox_dist_preds
        if self.training:
            return cls_logit, bbox_preds, bbox_dist_preds
        else:
            return cls_logit, bbox_preds



import numpy as np

from mmengine import load
from mmseg.datasets.transforms import LoadAnnotations as MMSeg_LoadAnnotations
from mmengine.registry import TRANSFORMS

from visaionlibrary.datasets.utils import decode_str2mask, SampleStatus


@TRANSFORMS.register_module()
class LoadSegAnnotations(MMSeg_LoadAnnotations):
    def _load_seg_map(self, results: dict) -> None:
        seg_map_path = results['seg_map_path']  # 标注json路径
        height, width = results['img_shape']

        if results['sample_type'] == SampleStatus.RAW:
            gt_semantic_seg = np.ones((height, width), dtype=np.uint8) * 255  # 255 表示为无标注
        elif results['sample_type'] == SampleStatus.BACKGROUND:
            gt_semantic_seg = np.zeros((height, width), dtype=np.uint8)
        elif results['sample_type'] == SampleStatus.FOREGROUND:
            masks = list()  # 用于存储一张图中的多个object的mask
            segmentations = load(seg_map_path)['segmentations']  # 读取json文件, 并获取segmentation字段
            for segmentation in segmentations:
                range_box = segmentation.get('rangeBox')  # 标注子框 x1, y1, w, h
                segmentation_map = segmentation.get('segmentationMap')  # 子框标注图
                label_name = segmentation.get('labelName')  # 获得子框对应的类别
                # for inference class not in metainfo, 推理时, 如果类别不在metainfo中, 则不进行读取, 认为是负样本
                if label_name in results['dataset_metainfo']['classes']:
                    # 获取mask_sub
                    label_index = results['dataset_metainfo']['classes'].index(label_name)  # 通过类别名称获得类别ID
                    range_box_width, range_box_height = range_box[2], range_box[3]  # 获得子框的宽, 高
                    mask_sub = decode_str2mask(segmentation_map, range_box_height, range_box_width)  # 获得子框的mask
                    mask_sub = np.where(mask_sub == 1, label_index, 0)  # 赋值对应labelIndex

                    # 获得mask
                    mask = np.zeros((height, width), dtype=np.uint8)  
                    # check rangebox and mask sub
                    if range_box[1] + range_box[3] >= height:
                        range_box[3] = height - range_box[1]
                    if range_box[0] + range_box[2] >= width:
                        range_box[2] = width - range_box[0]
                    mask_sub = mask_sub[:range_box[3], :range_box[2]]

                    mask[range_box[1]:range_box[1] + range_box[3], range_box[0]:range_box[0] + range_box[2]] = mask_sub
                    masks.append(mask)  # add mask image into mask_channels
            
            # merge masks
            masks = np.stack(masks, axis=0)  # 将多个mask合并

            # mask1  mask2  备注:(1)同一像素点, (2)0表示背景,2表示类别2,3表示类别3
            # 0      3  --> 情况1
            # 3      3  --> 情况2
            # 2      3  --> 情况3 (前端已经过滤掉了)
            gt_semantic_seg = np.max(masks, axis=0)

        # reduce zero_label
        if self.reduce_zero_label is None:
            self.reduce_zero_label = results['reduce_zero_label']
        assert self.reduce_zero_label == results['reduce_zero_label'], \
            'Initialize dataset with `reduce_zero_label` as ' \
            f'{results["reduce_zero_label"]} but when load annotation ' \
            f'the `reduce_zero_label` is {self.reduce_zero_label}'
        if self.reduce_zero_label:
            # avoid using underflow conversion
            gt_semantic_seg[gt_semantic_seg == 0] = 255
            gt_semantic_seg = gt_semantic_seg - 1
            gt_semantic_seg[gt_semantic_seg == 254] = 255
        # modify if custom classes
        if results.get('label_map', None) is not None:
            # Add deep copy to solve bug of repeatedly
            # replace `gt_semantic_seg`, which is reported in
            # https://github.com/open-mmlab/mmsegmentation/pull/1445/
            gt_semantic_seg_copy = gt_semantic_seg.copy()
            for old_id, new_id in results['label_map'].items():
                gt_semantic_seg[gt_semantic_seg_copy == old_id] = new_id
        results['gt_seg_map'] = gt_semantic_seg
        results['seg_fields'].append('gt_seg_map')




# Copyright (c) OpenMMLab. All rights reserved.
import random
from numbers import Number
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine.dist import barrier, broadcast, get_dist_info
from mmengine.logging import MessageHub
from mmengine.model import BaseDataPreprocessor, ImgDataPreprocessor
from mmengine.structures import PixelData
from mmengine.utils import is_seq_of
from mmengine.registry import MODELS as MMENGINE_MODELS
from torch import Tensor

from mmdet.models.utils import unfold_wo_center
from mmdet.models.utils.misc import samplelist_boxtype2tensor
from mmdet.registry import MODELS
from mmdet.structures import DetDataSample
from mmdet.structures.mask import BitmapMasks
from mmdet.utils import ConfigType

try:
    import skimage
except ImportError:
    skimage = None


@MMENGINE_MODELS.register_module()
class DetDataPreprocessor(ImgDataPreprocessor):
    """Image pre-processor for detection tasks.

    Comparing with the :class:`mmengine.ImgDataPreprocessor`,

    1. It supports batch augmentations.
    2. It will additionally append batch_input_shape and pad_shape
    to data_samples considering the object detection task.

    It provides the data pre-processing as follows

    - Collate and move data to the target device.
    - Pad inputs to the maximum size of current batch with defined
      ``pad_value``. The padding size can be divisible by a defined
      ``pad_size_divisor``
    - Stack inputs to batch_inputs.
    - Convert inputs from bgr to rgb if the shape of input is (3, H, W).
    - Normalize image with defined std and mean.
    - Do batch augmentations during training.

    Args:
        mean (Sequence[Number], optional): The pixel mean of R, G, B channels.
            Defaults to None.
        std (Sequence[Number], optional): The pixel standard deviation of
            R, G, B channels. Defaults to None.
        pad_size_divisor (int): The size of padded image should be
            divisible by ``pad_size_divisor``. Defaults to 1.
        pad_value (Number): The padded pixel value. Defaults to 0.
        pad_mask (bool): Whether to pad instance masks. Defaults to False.
        mask_pad_value (int): The padded pixel value for instance masks.
            Defaults to 0.
        pad_seg (bool): Whether to pad semantic segmentation maps.
            Defaults to False.
        seg_pad_value (int): The padded pixel value for semantic
            segmentation maps. Defaults to 255.
        bgr_to_rgb (bool): whether to convert image from BGR to RGB.
            Defaults to False.
        rgb_to_bgr (bool): whether to convert image from RGB to RGB.
            Defaults to False.
        boxtype2tensor (bool): Whether to keep the ``BaseBoxes`` type of
            bboxes data or not. Defaults to True.
        non_blocking (bool): Whether block current process
            when transferring data to device. Defaults to False.
        batch_augments (list[dict], optional): Batch-level augmentations
    """

    def __init__(self,
                 mean: Sequence[Number] = None,
                 std: Sequence[Number] = None,
                 pad_size_divisor: int = 1,
                 pad_value: Union[float, int] = 0,
                 pad_mask: bool = False,
                 mask_pad_value: int = 0,
                 pad_seg: bool = False,
                 seg_pad_value: int = 255,
                 grayscale_to_color: bool = False,
                 bgr_to_rgb: bool = False,
                 rgb_to_bgr: bool = False,
                 boxtype2tensor: bool = True,
                 non_blocking: Optional[bool] = False,
                 batch_augments: Optional[List[dict]] = None):
        super().__init__(
            mean=mean,
            std=std,
            pad_size_divisor=pad_size_divisor,
            pad_value=pad_value,
            bgr_to_rgb=bgr_to_rgb,
            rgb_to_bgr=rgb_to_bgr,
            non_blocking=non_blocking)
        if batch_augments is not None:
            self.batch_augments = nn.ModuleList(
                [MODELS.build(aug) for aug in batch_augments])
        else:
            self.batch_augments = None
        self.pad_mask = pad_mask
        self.mask_pad_value = mask_pad_value
        self.pad_seg = pad_seg
        self.seg_pad_value = seg_pad_value
        self.boxtype2tensor = boxtype2tensor
        self.grayscale_to_color = grayscale_to_color

    def forward(self, data: dict, training: bool = False) -> dict:
        """Perform normalization、padding and bgr2rgb conversion based on
        ``BaseDataPreprocessor``.

        Args:
            data (dict): Data sampled from dataloader.
            training (bool): Whether to enable training time augmentation.

        Returns:
            dict: Data in the same format as the model input.
        """
        batch_pad_shape = self._get_pad_shape(data)
        # convert grayscale to color if needed
        if self.grayscale_to_color:
            data['inputs'] = [
                torch.repeat_interleave(input_, 3, dim=0) if input_.size(0) == 1 else input_
                for input_ in data['inputs']
            ]

        data = super().forward(data=data, training=training)
        inputs, data_samples = data['inputs'], data['data_samples']

        if data_samples is not None:
            # NOTE the batched image size information may be useful, e.g.
            # in DETR, this is needed for the construction of masks, which is
            # then used for the transformer_head.
            batch_input_shape = tuple(inputs[0].size()[-2:])
            for data_sample, pad_shape in zip(data_samples, batch_pad_shape):
                data_sample.set_metainfo({
                    'batch_input_shape': batch_input_shape,
                    'pad_shape': pad_shape
                })

            if self.boxtype2tensor:
                samplelist_boxtype2tensor(data_samples)

            if self.pad_mask and training:
                self.pad_gt_masks(data_samples)

            if self.pad_seg and training:
                self.pad_gt_sem_seg(data_samples)

        if training and self.batch_augments is not None:
            for batch_aug in self.batch_augments:
                inputs, data_samples = batch_aug(inputs, data_samples)

        return {'inputs': inputs, 'data_samples': data_samples}

    def _get_pad_shape(self, data: dict) -> List[tuple]:
        """Get the pad_shape of each image based on data and
        pad_size_divisor."""
        _batch_inputs = data['inputs']
        # Process data with `pseudo_collate`.
        if is_seq_of(_batch_inputs, torch.Tensor):
            batch_pad_shape = []
            for ori_input in _batch_inputs:
                pad_h = int(
                    np.ceil(ori_input.shape[1] /
                            self.pad_size_divisor)) * self.pad_size_divisor
                pad_w = int(
                    np.ceil(ori_input.shape[2] /
                            self.pad_size_divisor)) * self.pad_size_divisor
                batch_pad_shape.append((pad_h, pad_w))
        # Process data with `default_collate`.
        elif isinstance(_batch_inputs, torch.Tensor):
            assert _batch_inputs.dim() == 4, (
                'The input of `ImgDataPreprocessor` should be a NCHW tensor '
                'or a list of tensor, but got a tensor with shape: '
                f'{_batch_inputs.shape}')
            pad_h = int(
                np.ceil(_batch_inputs.shape[1] /
                        self.pad_size_divisor)) * self.pad_size_divisor
            pad_w = int(
                np.ceil(_batch_inputs.shape[2] /
                        self.pad_size_divisor)) * self.pad_size_divisor
            batch_pad_shape = [(pad_h, pad_w)] * _batch_inputs.shape[0]
        else:
            raise TypeError('Output of `cast_data` should be a dict '
                            'or a tuple with inputs and data_samples, but got'
                            f'{type(data)}： {data}')
        return batch_pad_shape

    def pad_gt_masks(self,
                     batch_data_samples: Sequence[DetDataSample]) -> None:
        """Pad gt_masks to shape of batch_input_shape."""
        if 'masks' in batch_data_samples[0].gt_instances:
            for data_samples in batch_data_samples:
                masks = data_samples.gt_instances.masks
                data_samples.gt_instances.masks = masks.pad(
                    data_samples.batch_input_shape,
                    pad_val=self.mask_pad_value)

    def pad_gt_sem_seg(self,
                       batch_data_samples: Sequence[DetDataSample]) -> None:
        """Pad gt_sem_seg to shape of batch_input_shape."""
        if 'gt_sem_seg' in batch_data_samples[0]:
            for data_samples in batch_data_samples:
                gt_sem_seg = data_samples.gt_sem_seg.sem_seg
                h, w = gt_sem_seg.shape[-2:]
                pad_h, pad_w = data_samples.batch_input_shape
                gt_sem_seg = F.pad(
                    gt_sem_seg,
                    pad=(0, max(pad_w - w, 0), 0, max(pad_h - h, 0)),
                    mode='constant',
                    value=self.seg_pad_value)
                data_samples.gt_sem_seg = PixelData(sem_seg=gt_sem_seg)



from visaionlibrary.models.seg.encoder_decoder import EncoderDecoder
from visaionlibrary.models.seg.decode_head import BaseDecodeHead
from visaionlibrary.models.seg.hyh_unet_head import HYHUNetHead
from visaionlibrary.models.seg.regnet import RegNetVisaion

__all__ = ["EncoderDecoder", "BaseDecodeHead", "HYHUNetHead", "RegNetVisaion"]


from pathlib import Path
from types import SimpleNamespace
from typing import Dict, Union

from visaionlibrary.utils.yoloe.utils import (
    DEFAULT_CFG_DICT,
    IterableSimpleNamespace,
    yaml_load,
)

# Simplified keys for prediction only
CFG_FLOAT_KEYS = {"batch"}
CFG_FRACTION_KEYS = {"conf", "iou"}
CFG_INT_KEYS = {"max_det", "vid_stride", "line_width"}
CFG_BOOL_KEYS = {
    "save", "exist_ok", "verbose", "half", "dnn", "show",
    "save_txt", "save_conf", "save_crop", "save_frames",
    "show_labels", "show_conf", "visualize", "augment",
    "agnostic_nms", "retina_masks", "show_boxes",
}


def cfg2dict(cfg):
    """Converts a configuration object to a dictionary."""
    if isinstance(cfg, (str, Path)):
        cfg = yaml_load(cfg)
    elif isinstance(cfg, SimpleNamespace):
        cfg = vars(cfg)
    return cfg


def get_cfg(cfg: Union[str, Path, Dict, SimpleNamespace] = DEFAULT_CFG_DICT, overrides: Dict = None):
    """Load and merge configuration data from a file or dictionary, with optional overrides."""
    cfg = cfg2dict(cfg)

    # Merge overrides
    if overrides:
        overrides = cfg2dict(overrides)
        if "save_dir" not in cfg:
            overrides.pop("save_dir", None)
        cfg = {**cfg, **overrides}

    # Special handling for numeric project/name
    for k in "project", "name":
        if k in cfg and isinstance(cfg[k], (int, float)):
            cfg[k] = str(cfg[k])
    if cfg.get("name") == "model":
        cfg["name"] = cfg.get("model", "").split(".")[0]

    # Type checks (simplified)
    check_cfg(cfg)

    return IterableSimpleNamespace(**cfg)


def check_cfg(cfg, hard=False):
    """Checks configuration argument types and values (simplified version)."""
    for k, v in cfg.items():
        if v is not None:
            if k in CFG_FLOAT_KEYS and not isinstance(v, (int, float)):
                cfg[k] = float(v)
            elif k in CFG_FRACTION_KEYS:
                if not isinstance(v, (int, float)):
                    cfg[k] = v = float(v)
            elif k in CFG_INT_KEYS and not isinstance(v, int):
                cfg[k] = int(v)
            elif k in CFG_BOOL_KEYS and not isinstance(v, bool):
                cfg[k] = bool(v)



import random
from typing import Optional, Sequence, Tuple, Union
import math
import numpy as np

import mmcv
from mmcv.transforms.base import BaseTransform
from mmengine.registry import TRANSFORMS

Number = Union[int, float]


def _scale_size(
    size: Tuple[int, int],
    scale: Union[float, int, tuple],
) -> Tuple[int, int]:
    """Rescale a size by a ratio.

    Args:
        size (tuple[int]): (w, h).
        scale (float | tuple(float)): Scaling factor.

    Returns:
        tuple[int]: scaled size.
    """
    if isinstance(scale, (float, int)):
        scale = (scale, scale)
    w, h = size
    return int(w * float(scale[0]) + 0.5), int(h * float(scale[1]) + 0.5)


def rotate_point(x, y, cx, cy, angle_rad):
    x_new = (x - cx) * math.cos(angle_rad) - (y - cy) * math.sin(angle_rad) + cx
    y_new = (x - cx) * math.sin(angle_rad) + (y - cy) * math.cos(angle_rad) + cy
    return int(round(x_new, 0)), int(round(y_new, 0))


def rotate_region(x, y, width, height, cx, cy, angle_deg):
    angle = math.radians(angle_deg)
    # 构造原始矩阵的坐标矩阵
    X, Y = np.meshgrid(np.arange(width), np.arange(height))

    # 将坐标转换为以旋转中心为原点的坐标系
    X = X - (cx - x)
    Y = Y - (cy - y)

    # 应用旋转矩阵
    X_rotated = X * np.cos(angle) - Y * np.sin(angle)
    Y_rotated = X * np.sin(angle) + Y * np.cos(angle)

    # 将坐标转换回原始坐标系
    X_rotated += cx - x
    Y_rotated += cy - y

    # 将浮点型坐标四舍五入为整数型
    X_rotated = np.round(X_rotated).astype(int).flatten()
    Y_rotated = np.round(Y_rotated).astype(int).flatten()

    # 找到旋转后还在图像范围内的像素点并拉成直线
    condition = np.where(
        (X_rotated >= 0)
        & (X_rotated <= width - 1)
        & (Y_rotated >= 0)
        & (Y_rotated <= height - 1)
    )

    return X_rotated[condition], Y_rotated[condition]


@TRANSFORMS.register_module()
class RandomRescale(BaseTransform):
    """
    合并了resize和ALB的RandomScale的版本

    Required Keys:

    - seg_map_path: Path to the segmentation map file.  TODO:

    Modified Keys:


    Add Keys:
    - gt_seg_map(np.uint8): Ground truth segmentation map.  TODO:

    """

    def __init__(
        self,
        base_scale_factor: float = 1.0,
        scale_range: float = 0.1,
        prob: float = 0.5,
        keep_ratio: bool = True,
        backend: str = "cv2",
        interpolation="bilinear",
        clip_object_border: bool = True,
    ) -> None:
        assert 0 <= prob <= 1
        assert base_scale_factor > scale_range >= 0
        self.backend = backend
        self.interpolation = interpolation
        self.keep_ratio = keep_ratio
        self.clip_object_border = clip_object_border
        self.base_scale_factor = base_scale_factor
        self.scale_range = scale_range
        self.prob = prob

    def _resize_img(self, results: dict) -> None:
        """Resize images with ``results['scale']``."""
        if results.get("img", None) is not None:
            if self.keep_ratio:
                img, scale_factor = mmcv.imrescale(
                    results["img"],
                    results["scale"],
                    interpolation=self.interpolation,
                    return_scale=True,
                    backend=self.backend,
                )
                # the w_scale and h_scale has minor difference
                # a real fix should be done in the mmcv.imrescale in the future
                new_h, new_w = img.shape[:2]
                h, w = results["img"].shape[:2]
                w_scale = new_w / w
                h_scale = new_h / h
            else:
                img, w_scale, h_scale = mmcv.imresize(
                    results["img"],
                    results["scale"],
                    interpolation=self.interpolation,
                    return_scale=True,
                    backend=self.backend,
                )
            #  如果输入图像是灰度图，本来是[h, w, 1]这种size经过mmcv的resize就变成了[h, w]这种size，少了一维后面直接报错, by fufa
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=2)
            results["img"] = img
            results["img_shape"] = img.shape[:2]
            results["scale_factor"] = (w_scale, h_scale)
            results["keep_ratio"] = self.keep_ratio

    def _resize_bboxes(self, results: dict) -> None:
        """Resize bounding boxes with ``results['scale_factor']``."""
        if results.get("gt_bboxes", None) is not None:
            results["gt_bboxes"].rescale_(results["scale_factor"])
            if self.clip_object_border:
                results["gt_bboxes"].clip_(results["img_shape"])

    def _resize_masks(self, results: dict) -> None:
        """Resize masks with ``results['scale']``"""
        if results.get('gt_masks', None) is not None:
            if self.keep_ratio:
                results['gt_masks'] = results['gt_masks'].rescale(
                    results['scale'])
            else:
                results['gt_masks'] = results['gt_masks'].resize(
                    results['img_shape'])

    def _resize_seg(self, results: dict) -> None:
        """Resize semantic segmentation map with ``results['scale']``."""
        if results.get("gt_seg_map", None) is not None:
            if self.keep_ratio:
                gt_seg = mmcv.imrescale(
                    results["gt_seg_map"],
                    results["scale"],
                    interpolation="nearest",
                    backend=self.backend,
                )
            else:
                gt_seg = mmcv.imresize(
                    results["gt_seg_map"],
                    results["scale"],
                    interpolation="nearest",
                    backend=self.backend,
                )
            results["gt_seg_map"] = gt_seg

    def _resize_keypoints(self, results: dict) -> None:
        """Resize keypoints with ``results['scale_factor']``."""
        if results.get("gt_keypoints", None) is not None:
            keypoints = results["gt_keypoints"]
            keypoints[:, :, :2] = keypoints[:, :, :2] * np.array(
                results["scale_factor"]
            )
            if self.clip_object_border:
                keypoints[:, :, 0] = np.clip(
                    keypoints[:, :, 0], 0, results["img_shape"][1]
                )
                keypoints[:, :, 1] = np.clip(
                    keypoints[:, :, 1], 0, results["img_shape"][0]
                )
            results["gt_keypoints"] = keypoints

    def transform(self, results: dict) -> dict:
        img_shape = results["img"].shape[:2]
        if random.random() > self.prob:  # 说明直接进行固定比例的缩放
            results["scale"] = _scale_size(img_shape[::-1], self.base_scale_factor)  # type: ignore
        else:  # 需要进行随机比例缩放
            scale_factor = (
                (2 * random.random() * self.scale_range)
                - self.scale_range
                + self.base_scale_factor
            )
            results["scale"] = _scale_size(img_shape[::-1], scale_factor)  # type: ignore
        self._resize_img(results)
        self._resize_bboxes(results)
        self._resize_masks(results)
        self._resize_seg(results)
        self._resize_keypoints(results)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f"(scale={self.scale}, "
        repr_str += f"scale_factor={self.scale_factor}, "
        repr_str += f"keep_ratio={self.keep_ratio}, "
        repr_str += f"clip_object_border={self.clip_object_border}), "
        repr_str += f"backend={self.backend}), "
        repr_str += f"interpolation={self.interpolation})"
        return repr_str



import os

# Set ENV variables (place before imports)
if not os.environ.get("OMP_NUM_THREADS"):
    os.environ["OMP_NUM_THREADS"] = "1"  # default for reduced CPU utilization during training

from visaionlibrary.utils.yoloe.models import YOLO, YOLOE
from visaionlibrary.utils.yoloe.utils.checks import check_yolo as checks

__all__ = (
    "YOLO",
    "YOLOE",
    "checks",
)



from visaionlibrary.utils.yoloe.data.loaders import (
    LoadPilAndNumpy,
    SourceTypes,
)


def load_inference_source(source=None, batch=1, vid_stride=1, buffer=False):
    """
    Loads an inference source for object detection and applies necessary transformations.

    Args:
        source (str, Path, Tensor, PIL.Image, np.ndarray): The input source for inference.
        batch (int, optional): Batch size for dataloaders. Default is 1.
        vid_stride (int, optional): The frame interval for video sources. Default is 1.
        buffer (bool, optional): Determined whether stream frames will be buffered. Default is False.

    Returns:
        dataset (Dataset): A dataset object for the specified input source.
    """
    source, stream, screenshot, from_img, in_memory, tensor = source, False, False, True, False, False  
    source_type = source.source_type if in_memory else SourceTypes(stream, screenshot, from_img, tensor)

    # Dataloader
    dataset = LoadPilAndNumpy(source)

    # Attach source types to the dataset
    setattr(dataset, "source_type", source_type)

    return dataset



from typing import Dict
from mmdet.evaluation.metrics.coco_metric import CocoMetric
from mmengine.registry import METRICS

@METRICS.register_module()
class CocoMetricVisaion(CocoMetric):
    def compute_metrics(self, results: list) -> Dict[str, float]:
        eval_results = super().compute_metrics(results)

        # 新加内容, 澔哥要求只保留mAP, mAP_50
        only_keep_keys = ['bbox_mAP', 'bbox_mAP_50']
        eval_results = {k: v for k, v in eval_results.items() if k in only_keep_keys}
        return eval_results



from abc import abstractmethod
import clip
import mobileclip
import torch.nn as nn
from visaionlibrary.utils.yoloe.utils.torch_utils import smart_inference_mode
import torch

class TextModel(nn.Module):
    def __init__(self):
        super().__init__()
    
    @abstractmethod
    def tokenize(texts):
        pass
    
    @abstractmethod
    def encode_text(texts, dtype):
        pass

class CLIP(TextModel):
    def __init__(self, size, device):
        super().__init__()
        self.model = clip.load(size, device=device)[0]
        self.to(device)
        self.device = device
        self.eval()
    
    def tokenize(self, texts):
        return clip.tokenize(texts).to(self.device)
    
    @smart_inference_mode()
    def encode_text(self, texts, dtype=torch.float32):
        txt_feats = self.model.encode_text(texts).to(dtype)
        txt_feats = txt_feats / txt_feats.norm(p=2, dim=-1, keepdim=True)
        return txt_feats
        
class MobileCLIP(TextModel):
    
    config_size_map = {
        "s0": "s0",
        "s1": "s1",
        "s2": "s2",
        "b": "b",
        "blt": "b"
    }
    
    def __init__(self, size, device):
        super().__init__()
        config = self.config_size_map[size]
        self.model = mobileclip.create_model_and_transforms(f'mobileclip_{config}', pretrained=f'mobileclip_{size}.pt', device=device)[0]
        self.tokenizer = mobileclip.get_tokenizer(f'mobileclip_{config}')
        self.to(device)
        self.device = device
        self.eval()
    
    def tokenize(self, texts):
        text_tokens = self.tokenizer(texts).to(self.device)
        # max_len = text_tokens.argmax(dim=-1).max().item() + 1
        # text_tokens = text_tokens[..., :max_len]
        return text_tokens

    @smart_inference_mode()
    def encode_text(self, texts, dtype=torch.float32):
        text_features = self.model.encode_text(texts).to(dtype)
        text_features /= text_features.norm(p=2, dim=-1, keepdim=True)
        return text_features

def build_text_model(variant, device=None):
    # LOGGER.info(f"Build text model {variant}")
    base, size = variant.split(":")
    if base == 'clip':
        return CLIP(size, device)
    elif base == 'mobileclip':
        return MobileCLIP(size, device)
    else:
        print("Variant not found")
        assert(False)


import copy
from typing import Dict, List, Union, Tuple, Optional
import numpy as np
import random
import cv2
import torch

from mmcv.transforms.base import BaseTransform
from mmcv.transforms.utils import cache_randomness
from mmdet.datasets.transforms.transforms import RandomCrop
from mmdet.structures.bbox import autocast_box_type
from mmengine.registry import TRANSFORMS
from visaionlibrary.datasets.utils import SampleStatus


@TRANSFORMS.register_module()
class SegRegionSampler(BaseTransform):
    """Region Sampler for Train phase.
        通过在缺陷区域随机裁剪图像进行模型的训练

    
    Required Keys:

    - seg_map_path: Path to the segmentation map file.  TODO:
    
    Modified Keys:

    
    Add Keys:
    - gt_seg_map(np.uint8): Ground truth segmentation map.  TODO:


    Args:
        region_size: 小图的尺寸,(高, 宽)
        patience: 在正样本里面取负样本小图的时候,最多尝试的次数,如果设置成1就是随机裁剪,不管里面是否包含前景,默认为1
        pad_mode: pad填充模式
        centerness_target:目标是否基本在图像中央
        dual_mode:是否额外裁剪出一个样本来改善平移稳定性
    """
    def __init__(self,
                 region_size: int = 256,
                 patience: int = 3,
                 pad_mode: str = 'constant',
                 centerness_target: bool = False,
                 dual_mode: bool = False,
                 defect_ratio: float = 0.0) -> None:
        self._region_h = region_size
        self._region_w = region_size
        assert 1 <= patience <= 5  # 限定patience在[1,5]之间,太大就太慢了
        self._patience = patience
        assert pad_mode in ['constant', 'reflect', 'wrap']  # 这个pad_mode暂时是失效的
        self.centerness_target = centerness_target  # 缺陷和过检区域是否大部分在图像中央
        self.dual_mode = dual_mode
        self.defect_ratio = defect_ratio  # 缺陷区域的最小保留比例, 默认0

    def _crop_region_and_gt(self, img, gt_seg_map, gt_top_left_x, gt_top_left_y):
        """根据随机出来的gt_top_left_y和gt_top_left_x进行region和gt的裁剪"""
        gt_top_left_x = int(gt_top_left_x)
        gt_top_left_y = int(gt_top_left_y)

        # 这个是根据x和pad_w计算图像左边需要pad的宽度,当x > pad_w的时候,意味着左边不用pad,就得到的0,不然就需要pad
        pad_w_l = max(0, -gt_top_left_x)
        # 这个是根据x和pad_w计算图像左边需要pad的宽度,
        pad_w_r = max(0, gt_top_left_x + self._region_w - img.shape[1])
        # 高度方向的逻辑同上
        pad_h_t = max(0, -gt_top_left_y)
        pad_h_b = max(0, gt_top_left_y + self._region_h - img.shape[0])

        # 根据计算得到的需要pad的量处理原始图像
        padded_img = np.pad(img, ((pad_h_t, pad_h_b), (pad_w_l, pad_w_r), (0, 0)))
        padded_gt = np.pad(gt_seg_map, ((pad_h_t, pad_h_b), (pad_w_l, pad_w_r)))

        # region_top_left_x/y 计算的是在pad后的图像坐标系下,裁剪region时候左上角坐标
        region_top_left_x = gt_top_left_x + pad_w_l
        region_top_left_y = gt_top_left_y + pad_h_t
        region = padded_img[region_top_left_y:region_top_left_y + self._region_h,
                            region_top_left_x:region_top_left_x + self._region_w, :]
        gt = padded_gt[region_top_left_y:region_top_left_y + self._region_h,
                       region_top_left_x:region_top_left_x + self._region_w]
        return region, gt

    def transform(self, results: dict) -> dict:
        """裁剪小图,根据其中的crop_mode来指定具体裁哪,是只裁前景位置还是裁背景位置

        Args:
            results(dict): Result dict from
                :class:`BaseTransform`

        Returns:
            dict: 返回裁剪后的字典.

        """
        sample_type = results['sample_type']    # 数据集中第i个数据的类型
        gt_seg_map = results['gt_seg_map']    # 获取标注图片
        img = results['img']                 # 获取图片
        # 获取shape,要直接从img计算而不是直接拿results['img_shape'],因为前面的增广过程可能会修改size
        img_h, img_w = img.shape[:2]

        # 1. 要么是sampler用错了,不是randomNG类的导致没这个字段
        # 2. 要么是dataset初始化读取图像类型的时候错误使用了后续的transform,在sampler之前就调用了本函数,也没这个字段
        if results.get('crop_mode', None) is None:
            # 但是这两种情况,都会有sample_type,因为就算是情况2也刚经过LoadAnn,所以根据type指定mode,就是退化了而已
            crop_mode = "label" if sample_type == SampleStatus.FOREGROUND else "background"
        else:
            crop_mode = results['crop_mode']  # 裁剪小图的模式,label:只裁有标注的地方,background:尽可能只在背景进行裁剪

        if gt_seg_map.max() > 0:  # 确定是缺陷样本
            if crop_mode == 'label':  # 正常裁剪正样本小图
                # 正样本就需要先pad,避免过小的图像出现,或者出现太过靠近边界的缺陷导致过多的pad影响训练
                pad_h = int(max(self._region_h - img_h, 0))
                pad_w = int(max(self._region_w - img_w, 0))
                pad_h_t = random.randint(0, pad_h)
                pad_h_b = pad_h - pad_h_t
                pad_w_l = random.randint(0, pad_w)
                pad_w_r = pad_w - pad_w_l

                img = np.pad(img, ((pad_h_t, pad_h_b), (pad_w_l, pad_w_r), (0, 0)))
                gt_seg_map = np.pad(gt_seg_map, ((pad_h_t, pad_h_b), (pad_w_l, pad_w_r)))  # gt 同样的处理
                img_h, img_w = img.shape[:2]  # 重新计算尺寸
                if self.defect_ratio > 0:
                    contours, _ = cv2.findContours(gt_seg_map, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
                    contour = random.choice(contours)
                    tmp_map = np.zeros_like(gt_seg_map)
                    cv2.drawContours(tmp_map, [contour], 0, 1, -1)
                    defect_ys, defect_xs = np.nonzero(tmp_map)
                    defect_area = len(defect_xs)  # 缺陷区域面积
                    max_try = 20  # 最多尝试20次,如果都找不到,就跳出循环
                    for _ in range(max_try):
                        defect_y, defect_x = random.choice(list(zip(defect_ys, defect_xs)))
                        # 随机化该缺陷点在_gt_h*_gt_w范围内的相对坐标 从0波动到self._gt_h - 1这个范围
                        gt_top_left_y = defect_y.item() - random.randint(0, self._region_h - 1)
                        gt_top_left_x = defect_x.item() - random.randint(0, self._region_w - 1)

                        # 如果都在中央,就可以随机裁剪不做限制,因为小图中一定包含大量原始图像信息
                        # 如果目标不都是在中央,那正样本就需要进行过界判定,不能出图像的范围,不然可能导致裁剪质量下降
                        if not self.centerness_target:
                            gt_top_left_y = max(gt_top_left_y, 0)
                            gt_top_left_x = max(gt_top_left_x, 0)
                            gt_top_left_y = min(gt_top_left_y, img_h - self._region_h)  # 这个就是说y不能超过计算得到的最大值
                            gt_top_left_x = min(gt_top_left_x, img_w - self._region_w)  # x方向上的逻辑与y方向相同

                        region, gt = self._crop_region_and_gt(img, gt_seg_map, gt_top_left_x, gt_top_left_y)
                        if len(np.nonzero(gt)[0]) >= defect_area * self.defect_ratio:
                            break

                else:
                    defect_ys, defect_xs = np.nonzero(gt_seg_map)
                    defect_y, defect_x = random.choice(list(zip(defect_ys, defect_xs)))
                    # 随机化该缺陷点在_gt_h*_gt_w范围内的相对坐标 从0波动到self._gt_h - 1这个范围
                    gt_top_left_y = defect_y.item() - random.randint(0, self._region_h - 1)
                    gt_top_left_x = defect_x.item() - random.randint(0, self._region_w - 1)

                    # 如果都在中央,就可以随机裁剪不做限制,因为小图中一定包含大量原始图像信息
                    # 如果目标不都是在中央,那正样本就需要进行过界判定,不能出图像的范围,不然可能导致裁剪质量下降
                    if not self.centerness_target:
                        gt_top_left_y = max(gt_top_left_y, 0)
                        gt_top_left_x = max(gt_top_left_x, 0)
                        gt_top_left_y = min(gt_top_left_y, img_h - self._region_h)  # 这个就是说y不能超过计算得到的最大值
                        gt_top_left_x = min(gt_top_left_x, img_w - self._region_w)  # x方向上的逻辑与y方向相同

                    region, gt = self._crop_region_and_gt(img, gt_seg_map, gt_top_left_x, gt_top_left_y)

            else:  # 从正样本裁剪负样本小图
                gt_top_left_x, gt_top_left_y = None, None  # 事先定义几个None,来确保出了for之后,region和gt还能访问到
                region, gt = None, None
                for _ in range(self._patience):  # 最多尝试patience次,次数到了之后就算还是裁剪到了前景区域也不管了
                    gt_top_left_y = random.randint(0, img_h) - 0.5 * self._region_h  # gt_top_left_x/y是裁剪区域左上角点的坐标
                    gt_top_left_x = random.randint(0, img_w) - 0.5 * self._region_w
                    region, gt = self._crop_region_and_gt(img, gt_seg_map, gt_top_left_x, gt_top_left_y)
                    if gt.max() == 0:  # 如果裁到了负样本,就跳出循环
                        break

        else:
            # 背景样本的话,就是只限制裁剪小图的中心点不出原始图像的范围
            gt_top_left_y = random.randint(0, img_h) - 0.5 * self._region_h  # gt_top_left_x/y是裁剪区域左上角点的坐标
            gt_top_left_x = random.randint(0, img_w) - 0.5 * self._region_w
            region, gt = self._crop_region_and_gt(img, gt_seg_map, gt_top_left_x, gt_top_left_y)

        results['img'] = np.ascontiguousarray(region)   # 将region转换为连续的内存块
        results['img_shape'] = region.shape[:2]
        results['gt_seg_map'] = np.ascontiguousarray(gt)  # 将gt转换为连续的内存块
        results['coord'] = [gt_top_left_x, gt_top_left_y, self._region_h, self._region_w]

        if self.dual_mode:
            # Generating the dual sample, translated within 25% of crop size.
            trans_range = int(self._region_h / 4.)
            dual_trans_x = random.randint(1, trans_range) if random.random() > 0.5 else -random.randint(1, trans_range)
            dual_trans_y = random.randint(1, trans_range) if random.random() > 0.5 else -random.randint(1, trans_range)

            dual_gt_top_left_x = gt_top_left_x + dual_trans_x
            dual_gt_top_left_y = gt_top_left_y + dual_trans_y
            # temporary ignore the out-range problem
            dual_region, dual_gt = self._crop_region_and_gt(img, gt_seg_map, dual_gt_top_left_x, dual_gt_top_left_y)

            results['dual_img'] = dual_region
            results['dual_gt_seg_map'] = dual_gt

            # 对偶样本不进行旋转,确保角度较小情况下统计对应关系
            img_h, img_w = results['img_shape']
            # Find the corresponding area before rotation
            region_orig = [max(0, dual_trans_x), max(0, dual_trans_y),
                           min(img_w, img_w + dual_trans_x), min(img_h, img_h + dual_trans_y)]
            region_dual = [max(0, -dual_trans_x), max(0, -dual_trans_y),
                           min(img_w, img_w - dual_trans_x), min(img_h, img_h - dual_trans_y)]

            region_w = region_orig[2] - region_orig[0]  # 重叠区域的宽高
            region_h = region_orig[3] - region_orig[1]

            x_grid_orig, y_grid_orig = np.meshgrid(np.arange(region_orig[0], region_orig[0] + region_w),
                                                   np.arange(region_orig[1], region_orig[1] + region_h))
            x_grid_dual, y_grid_dual = np.meshgrid(np.arange(region_dual[0], region_dual[0] + region_w),
                                                   np.arange(region_dual[1], region_dual[1] + region_h))
            # 这里存储的就是两张图像里边信息对应的一系列坐标点,都是按相同的顺序存储的
            results['overlap_region'] = (x_grid_orig.flatten(), y_grid_orig.flatten(),
                                         x_grid_dual.flatten(), y_grid_dual.flatten())

        return results

@TRANSFORMS.register_module()
class DetRegionSampler(RandomCrop):
    """Crop small img from oring image
    1. select a bbox from image
    2. bbox random in new img coord, only make sure hole box is inside.
    3. if box is large than crop image, crop image inside box
    4. Pad extra area

    off_rate: close

    """
    def __init__(self,
                 crop_size: tuple,  # (w, h)
                 crop_type: str = 'absolute',
                 allow_negative_crop: bool = True,
                 recompute_bbox: bool = False,
                 bbox_clip_border: bool = True,
                 enlarge_img: bool = True,
                 off_rate: float = 0.0,
                 pad_val: Union[Union[int, float], dict] = 0,
                 padding_mode: str = 'constant'
                 ) -> None:
        super(DetRegionSampler, self).__init__(
            crop_size=crop_size,
            crop_type=crop_type,
            allow_negative_crop=allow_negative_crop,
            recompute_bbox=recompute_bbox,
            bbox_clip_border=bbox_clip_border,
        )
        assert self.crop_size[0] > 0 and self.crop_size[1] > 0
        self.pad_val = pad_val
        self.padding_mode = padding_mode
        self.enlarge_img = enlarge_img
        self.off_rate = off_rate

    def _crop_data(self, results: dict, allow_negative_crop: bool) -> Union[dict, None]:
        """Function to randomly crop images, bounding boxes, masks, semantic
        segmentation maps.

        Args:
            results (dict): Result dict from loading pipeline.
            crop_size (Tuple[int, int]): Expected absolute size after
                cropping, (h, w).
            allow_negative_crop (bool): Whether to allow a crop that does not
                contain any bbox area.

        Returns:
            results (Union[dict, None]): Randomly cropped results, 'img_shape'
                key in result dict is updated according to crop size. None will
                be returned when there is no valid bbox after cropping.
        """
        img = results['img']
        margin_h = img.shape[0] - self.crop_size[1]
        margin_w = img.shape[1] - self.crop_size[0]
        offset_h, offset_w = self._rand_offset((margin_h, margin_w), results, img.shape[:2])  # by ldf
        crop_y1, crop_y2 = offset_h, offset_h + self.crop_size[1]
        crop_x1, crop_x2 = offset_w, offset_w + self.crop_size[0]

        # Record the homography matrix for the RandomCrop
        homography_matrix = np.array(
            [[1, 0, -offset_w], [0, 1, -offset_h], [0, 0, 1]],
            dtype=np.float32)
        if results.get('homography_matrix', None) is None:
            results['homography_matrix'] = homography_matrix
        else:
            results['homography_matrix'] = homography_matrix @ results[
                'homography_matrix']

        # crop and pad the image
        img = self._crop_and_pad_img(img, crop_x1, crop_y1, crop_x2, crop_y2)  # by ldf
        img_shape = img.shape
        results['img'] = img
        results['img_shape'] = img_shape[:2]

        # crop bboxes accordingly and clip to the image boundary
        if results.get('gt_bboxes', None) is not None:
            bboxes = results['gt_bboxes']
            bboxes.translate_([-offset_w, -offset_h])
            if self.bbox_clip_border:
                bboxes.clip_(img_shape[:2])
            valid_inds = bboxes.is_inside(img_shape[:2]).numpy()
            # If the crop does not contain any gt-bbox area and
            # allow_negative_crop is False, skip this image.
            if (not valid_inds.any() and not allow_negative_crop):
                return None

            results['gt_bboxes'] = bboxes[valid_inds]

            if results.get('gt_ignore_flags', None) is not None:
                results['gt_ignore_flags'] = \
                    results['gt_ignore_flags'][valid_inds]

            if results.get('gt_bboxes_labels', None) is not None:
                results['gt_bboxes_labels'] = \
                    results['gt_bboxes_labels'][valid_inds]

            if results.get('gt_masks', None) is not None:
                gt_masks = results['gt_masks'][valid_inds.nonzero()[0]]
                # translate the masks first, then crop
                gt_masks = gt_masks.translate((gt_masks.height, max(crop_x2, gt_masks.width)-min(crop_x1, 0)), -offset_w, direction="horizontal")  # translate in the horizontal
                gt_masks = gt_masks.translate((max(crop_y2, gt_masks.height)-min(crop_y1, 0), gt_masks.width), -offset_h, direction="vertical")  # translate in the vertical
                gt_masks = gt_masks.crop(np.asarray([0, 0, img_shape[1], img_shape[0]]))
                results['gt_masks'] = gt_masks

                if self.recompute_bbox:
                    results['gt_bboxes'] = results['gt_masks'].get_bboxes(
                        type(results['gt_bboxes']))

        # crop semantic seg
        if results.get('gt_seg_map', None) is not None:
            raise NotImplementedError("Segmentation map is not supported yet")


        return results

    def _crop_and_pad_img(self, img, x1, y1, x2, y2):
        """

        Args:
            img:
            crop_coords: x1, y1, x2, y2,

        Returns:

        """
        # TODO: more pad mode
        h, w = img.shape[:2]
        pad_size = [self.crop_size[1], self.crop_size[0]]
        if len(img.shape) == 3:
            pad_size.append(img.shape[2])
        if self.pad_val is not None:
            if isinstance(self.pad_val, int):
                pad_img = np.ones(pad_size, dtype=np.uint8) * self.pad_val
            elif isinstance(self.pad_val, dict):
                pad_img = np.ones(pad_size, dtype=np.uint8) * self.pad_val["img"]

            pad_img[max(0-y1, 0): min(y2-y1, h-y1), max(0-x1, 0): min(x2-x1, w-x1), ...] = \
                img[max(y1, 0): min(y2, h), max(x1, 0): min(x2, w), ...]
        else:
            raise ValueError("Plz input pad val, other pad mode not implementation yet")

        return pad_img

    @cache_randomness
    def _rand_offset(self, margin: Tuple[int, int], results: dict, img_size: Tuple[int, int]) -> Tuple[int, int]:
        """Randomly generate crop offset.

        Args:
            margin (Tuple[int, int]): The upper bound for the offset generated
                randomly.

        Returns:
            Tuple[int, int]: The random offset for the crop.
        """
        margin_h, margin_w = margin  # img size - crop size
        bboxes = results['gt_bboxes']  # (x1, y1, x2, y2)
        labels = results['gt_bboxes_labels']
        assert len(bboxes) == len(labels)
        if len(bboxes) > 0:
            bbox = bboxes[np.random.randint(0, len(bboxes))]  # random bbox
            x1, y1, x2, y2 = bbox.tensor[0]
            # ensure box in img
            x1, y1 = int(max(x1, 0)), int(max(y1, 0))
            x2, y2 = int(min(x2, img_size[1]-1)), int(min(y2, img_size[0]-1))
            if np.abs(x1 + self.crop_size[0] - x2) <= 1:
                offset_w = x1
            else:
                if self.enlarge_img:
                    w = int(img_size[1] * (1+self.off_rate))
                    if self.crop_size[0] < w:
                        if self.crop_size[0] > x2-x1:
                            offset_w = np.random.randint(max(x2-self.crop_size[0], 0), min(w-self.crop_size[0], x1)+1)
                        else:
                            offset_w = np.random.randint(x1, x2-self.crop_size[0]+1)
                    elif self.crop_size[0] == w:
                        offset_w = 0
                    else:
                        offset_w = np.random.randint(w-self.crop_size[0], 1)
                else:
                    offset_w = np.random.randint(min(x1, x2-self.crop_size[0]), max(x1, x2-self.crop_size[0])+1)
            if np.abs(y1 + self.crop_size[1] - y2) <= 1:
                offset_h = y1
            else:
                if self.enlarge_img:
                    h = int(img_size[0] * (1+self.off_rate))
                    if self.crop_size[1] < h:
                        if self.crop_size[1] > y2-y1:
                            offset_h = np.random.randint(max(y2-self.crop_size[1], 0), min(h-self.crop_size[1], y1)+1)
                        else:
                            offset_h = np.random.randint(y1, y2 - self.crop_size[1]+1)
                    elif self.crop_size[1] == h:
                        offset_h = 0
                    else:
                        offset_h = np.random.randint(h-self.crop_size[1], 1)
                else:
                    offset_h = np.random.randint(min(y1, y2-self.crop_size[1]), max(y1, y2-self.crop_size[1])+1)
        else:
            offset_h = np.random.randint(min(0, margin_h), max(0, margin_h)+1)
            offset_w = np.random.randint(min(0, margin_w), max(0, margin_w)+1)

        return offset_h, offset_w

    @autocast_box_type()
    def transform(self, results: dict) -> Union[dict, None]:
        """Transform function to randomly crop images, bounding boxes, masks,
        semantic segmentation maps.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            results (Union[dict, None]): Randomly cropped results, 'img_shape'
                key in result dict is updated according to crop size. None will
                be returned when there is no valid bbox after cropping.
        """
        results = self._crop_data(results, self.allow_negative_crop)
        return results


import cv2
import numpy as np
import torch
from torch.nn import functional as F

from visaionlibrary.utils.yoloe.utils.ops import xywh2xyxy


class LetterBox:
    """
    Resize image and padding for detection, instance segmentation, pose.

    This class resizes and pads images to a specified shape while preserving aspect ratio. It also updates
    corresponding labels and bounding boxes.

    Attributes:
        new_shape (tuple): Target shape (height, width) for resizing.
        auto (bool): Whether to use minimum rectangle.
        scaleFill (bool): Whether to stretch the image to new_shape.
        scaleup (bool): Whether to allow scaling up. If False, only scale down.
        stride (int): Stride for rounding padding.
        center (bool): Whether to center the image or align to top-left.

    Methods:
        __call__: Resize and pad image, update labels and bounding boxes.

    Examples:
        >>> transform = LetterBox(new_shape=(640, 640))
        >>> result = transform(labels)
        >>> resized_img = result["img"]
        >>> updated_instances = result["instances"]
    """

    def __init__(self, new_shape=(640, 640), auto=False, scaleFill=False, scaleup=True, center=True, stride=32):
        """
        Initialize LetterBox object for resizing and padding images.

        This class is designed to resize and pad images for object detection, instance segmentation, and pose estimation
        tasks. It supports various resizing modes including auto-sizing, scale-fill, and letterboxing.

        Args:
            new_shape (Tuple[int, int]): Target size (height, width) for the resized image.
            auto (bool): If True, use minimum rectangle to resize. If False, use new_shape directly.
            scaleFill (bool): If True, stretch the image to new_shape without padding.
            scaleup (bool): If True, allow scaling up. If False, only scale down.
            center (bool): If True, center the placed image. If False, place image in top-left corner.
            stride (int): Stride of the model (e.g., 32 for YOLOv5).

        Attributes:
            new_shape (Tuple[int, int]): Target size for the resized image.
            auto (bool): Flag for using minimum rectangle resizing.
            scaleFill (bool): Flag for stretching image without padding.
            scaleup (bool): Flag for allowing upscaling.
            stride (int): Stride value for ensuring image size is divisible by stride.

        Examples:
            >>> letterbox = LetterBox(new_shape=(640, 640), auto=False, scaleFill=False, scaleup=True, stride=32)
            >>> resized_img = letterbox(original_img)
        """
        self.new_shape = new_shape
        self.auto = auto
        self.scaleFill = scaleFill
        self.scaleup = scaleup
        self.stride = stride
        self.center = center  # Put the image in the middle or top-left

    def __call__(self, labels=None, image=None):
        """
        Resizes and pads an image for object detection, instance segmentation, or pose estimation tasks.

        This method applies letterboxing to the input image, which involves resizing the image while maintaining its
        aspect ratio and adding padding to fit the new shape. It also updates any associated labels accordingly.

        Args:
            labels (Dict | None): A dictionary containing image data and associated labels, or empty dict if None.
            image (np.ndarray | None): The input image as a numpy array. If None, the image is taken from 'labels'.

        Returns:
            (Dict | Tuple): If 'labels' is provided, returns an updated dictionary with the resized and padded image,
                updated labels, and additional metadata. If 'labels' is empty, returns a tuple containing the resized
                and padded image, and a tuple of (ratio, (left_pad, top_pad)).

        Examples:
            >>> letterbox = LetterBox(new_shape=(640, 640))
            >>> result = letterbox(labels={"img": np.zeros((480, 640, 3)), "instances": Instances(...)})
            >>> resized_img = result["img"]
            >>> updated_instances = result["instances"]
        """
        if labels is None:
            labels = {}
        img = labels.get("img") if image is None else image
        shape = img.shape[:2]  # current shape [height, width]
        new_shape = labels.pop("rect_shape", self.new_shape)
        if isinstance(new_shape, int):
            new_shape = (new_shape, new_shape)

        # Scale ratio (new / old)
        r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
        if not self.scaleup:  # only scale down, do not scale up (for better val mAP)
            r = min(r, 1.0)

        # Compute padding
        ratio = r, r  # width, height ratios
        new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
        dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
        if self.auto:  # minimum rectangle
            dw, dh = np.mod(dw, self.stride), np.mod(dh, self.stride)  # wh padding
        elif self.scaleFill:  # stretch
            dw, dh = 0.0, 0.0
            new_unpad = (new_shape[1], new_shape[0])
            ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

        if self.center:
            dw /= 2  # divide padding into 2 sides
            dh /= 2

        if shape[::-1] != new_unpad:  # resize
            img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
        top, bottom = int(round(dh - 0.1)) if self.center else 0, int(round(dh + 0.1))
        left, right = int(round(dw - 0.1)) if self.center else 0, int(round(dw + 0.1))
        img = cv2.copyMakeBorder(
            img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114)
        )  # add border
        if labels.get("ratio_pad"):
            labels["ratio_pad"] = (labels["ratio_pad"], (left, top))  # for evaluation

        if len(labels):
            labels = self._update_labels(labels, ratio, left, top)
            labels["img"] = img
            labels["resized_shape"] = new_shape
            return labels
        else:
            return img

    def _update_labels(self, labels, ratio, padw, padh):
        """
        Updates labels after applying letterboxing to an image.

        This method modifies the bounding box coordinates of instances in the labels
        to account for resizing and padding applied during letterboxing.

        Args:
            labels (Dict): A dictionary containing image labels and instances.
            ratio (Tuple[float, float]): Scaling ratios (width, height) applied to the image.
            padw (float): Padding width added to the image.
            padh (float): Padding height added to the image.

        Returns:
            (Dict): Updated labels dictionary with modified instance coordinates.

        Examples:
            >>> letterbox = LetterBox(new_shape=(640, 640))
            >>> labels = {"instances": Instances(...)}
            >>> ratio = (0.5, 0.5)
            >>> padw, padh = 10, 20
            >>> updated_labels = letterbox._update_labels(labels, ratio, padw, padh)
        """
        labels["instances"].convert_bbox(format="xyxy")
        labels["instances"].denormalize(*labels["img"].shape[:2][::-1])
        labels["instances"].scale(*ratio)
        labels["instances"].add_padding(padw, padh)
        return labels


class LoadVisualPrompt:
    def __init__(self, nc, augment):
        self.nc = nc
        self.min_interval = 5
        self.augment = augment
        self.scale_factor = 1/8
    
    def make_mask(self, boxes, h, w):
        x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(n,1,1)
        r = torch.arange(w)[None, None, :]  # rows shape(1,1,w)
        c = torch.arange(h)[None, :, None]  # cols shape(1,h,1)

        return ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))
    
    def __call__(self, labels):
        imgsz = labels["img"].shape[1:]
        masksz = (int(imgsz[0] * self.scale_factor), int(imgsz[1] * self.scale_factor))
        if "bboxes" in labels:
            bboxes = labels["bboxes"]
            bboxes = xywh2xyxy(bboxes) * torch.tensor(masksz)[[1, 0, 1, 0]]  # target boxes
            masks = self.make_mask(bboxes, *masksz).float()
        elif "masks" in labels:
            assert(not self.augment)
            masks = F.interpolate(torch.from_numpy(labels["masks"]).unsqueeze(1), 
                                  masksz, mode="nearest").squeeze(1).float()
        else:
            raise ValueError("LoadVisualPrompt must have bboxes or masks in the label")

        cls = labels["cls"].squeeze(-1).to(torch.int)
        if len(cls):
            visuals = torch.zeros(cls.max() + 1, *masksz)
            for idx, mask in zip(cls, masks):
                visuals[idx] = torch.logical_or(visuals[idx], mask)
        else:
            visuals = torch.zeros(0, *masksz)
        # visuals[0] = masks[random.choice(range(len(masks)))]
        labels["visuals"] = visuals
        return labels



import os
import os.path as osp
import cv2
import json
import torch
import numpy as np
import networkx as nx
from skimage import morphology
from pathlib import Path
from scipy.ndimage import label as sclabel
from scipy.ndimage import mean as scmean
from collections import OrderedDict
from prettytable import PrettyTable
from typing import Dict, List, Optional, Sequence, Union

from mmengine.evaluator import BaseMetric
from mmengine.logging import MMLogger, print_log
from mmseg.evaluation import IoUMetric as MMSeg_IoUMetric
from mmengine.registry import METRICS


@METRICS.register_module()
class IoUMetric(MMSeg_IoUMetric):
    """detail ref mmseg.evaluation.IOUMetric
    """

    def compute_metrics(self, results: list) -> Dict[str, float]:
        logger: MMLogger = MMLogger.get_current_instance()
        if self.format_only:
            logger.info(f'results are saved to {osp.dirname(self.output_dir)}')
            return OrderedDict()
        # convert list of tuples to tuple of lists, e.g.
        # [(A_1, B_1, C_1, D_1), ...,  (A_n, B_n, C_n, D_n)] to
        # ([A_1, ..., A_n], ..., [D_1, ..., D_n])
        results = tuple(zip(*results))
        assert len(results) == 4

        total_area_intersect = sum(results[0])[1:]
        total_area_union = sum(results[1])[1:]
        total_area_pred_label = sum(results[2])[1:]
        total_area_label = sum(results[3])[1:]
        ret_metrics = self.total_area_to_metrics(
            total_area_intersect, total_area_union, total_area_pred_label,
            total_area_label, self.metrics, self.nan_to_num, self.beta)

        class_names = self.dataset_meta['classes'][1:]

        # summary table
        ret_metrics_summary = OrderedDict({
            ret_metric: np.round(np.nanmean(ret_metric_value) * 100, 2)
            for ret_metric, ret_metric_value in ret_metrics.items()
        })
        metrics = dict()
        for key, val in ret_metrics_summary.items():
            if key == 'IoU':
            #     metrics[key] = val
            # else:
                metrics['m' + key] = val

        # each class table
        ret_metrics.pop('aAcc', None)
        ret_metrics_class = OrderedDict({
            ret_metric: np.round(ret_metric_value * 100, 2)
            for ret_metric, ret_metric_value in ret_metrics.items()
        })

        # Add classes acc, iou into metrics [add by ldf]
        for key, val in ret_metrics_class.items():
            for i, class_name in enumerate(class_names):
                metrics[key + '_' + class_name] = round(float(val[i]), 2)

        ret_metrics_class.update({'Class': class_names})
        ret_metrics_class.move_to_end('Class', last=False)
        class_table_data = PrettyTable()
        for key, val in ret_metrics_class.items():
            class_table_data.add_column(key, val)

        print_log('per class results:', logger)
        print_log('\n' + class_table_data.get_string(), logger=logger)

        return metrics

    @staticmethod
    def total_area_to_metrics(total_area_intersect: np.ndarray,
                              total_area_union: np.ndarray,
                              total_area_pred_label: np.ndarray,
                              total_area_label: np.ndarray,
                              metrics: List[str] = ['mIoU'],
                              nan_to_num: Optional[int] = None,
                              beta: int = 1):

        def f_score(precision, recall, beta=1):
            """calculate the f-score value.

            Args:
                precision (float | torch.Tensor): The precision value.
                recall (float | torch.Tensor): The recall value.
                beta (int): Determines the weight of recall in the combined
                    score. Default: 1.

            Returns:
                [torch.tensor]: The f-score value.
            """
            score = (1 + beta**2) * (precision * recall) / (
                (beta**2 * precision) + recall)
            return score

        if isinstance(metrics, str):
            metrics = [metrics]
        allowed_metrics = ['mIoU', 'mDice', 'mFscore']
        if not set(metrics).issubset(set(allowed_metrics)):
            raise KeyError(f'metrics {metrics} is not supported')

        all_acc = total_area_intersect.sum() / total_area_label.sum()
        ret_metrics = OrderedDict({'aAcc': all_acc})
        for metric in metrics:
            if metric == 'mIoU':
                iou = total_area_intersect / total_area_union
                # acc = total_area_intersect / total_area_label
                ret_metrics['IoU'] = iou
                # ret_metrics['Acc'] = acc
            elif metric == 'mDice':
                dice = 2 * total_area_intersect / (
                    total_area_pred_label + total_area_label)
                # acc = total_area_intersect / total_area_label
                ret_metrics['Dice'] = dice
                # ret_metrics['Acc'] = acc
            elif metric == 'mFscore':
                precision = total_area_intersect / total_area_pred_label
                recall = total_area_intersect / total_area_label
                f_value = torch.tensor([
                    f_score(x[0], x[1], beta) for x in zip(precision, recall)
                ])
                ret_metrics['Fscore'] = f_value
                ret_metrics['Precision'] = precision
                ret_metrics['Recall'] = recall

        ret_metrics = {
            metric: value.numpy()
            for metric, value in ret_metrics.items()
        }
        if nan_to_num is not None:
            ret_metrics = OrderedDict({
                metric: np.nan_to_num(metric_value, nan=nan_to_num)
                for metric, metric_value in ret_metrics.items()
            })
        return ret_metrics


@METRICS.register_module()
class VisaionMetric(BaseMetric):
    """Visaion evaluation metric. 基于无向图分析，改进多类情况下的计算过程

    """
    def __init__(self,
                 ignore_index: int = 255,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 iou_threshold: float = 0.5,
                 score_threshold: float = 0.5,
                 min_area: int = 20,
                 save_name: str = None,
                 no_argmax=False
                 ) -> None:
        super().__init__(collect_device=collect_device, prefix=prefix)

        self.ignore_index = ignore_index
        self.iou_threshold = iou_threshold
        self.score_threshold = score_threshold
        self.min_area = min_area
        self.overall_info_array = None
        self.save_name = save_name
        self.no_argmax = no_argmax

    def compute_region_paring(self,
                              pred_logit: torch.tensor,
                              pred_label: torch.tensor,
                              label: torch.tensor
                              ):
        """Calculate Confusion Matrix.
            Args:
                pred_logit (torch.tensor): Prediction origin after activation. The shape is (numClass, H, W).
                pred_label (torch.tensor): Prediction results
                label (torch.tensor): Ground truth segmentation map. The shape is (H, W).
            Returns:
                multiclass_info_list (list)
        """

        pred_activated = pred_logit.max(dim=0, keepdim=True)[0].squeeze()
        pred_activated = pred_activated.cpu().numpy()  # (h, w)
        mask = (label != self.ignore_index)
        pred_label = (pred_label * mask).cpu().numpy()
        label = (label * mask).cpu().numpy()
        num_classes = pred_logit.shape[0]

        # 得到无向图，和对应GT的数量
        graph, gt_nums = self.build_graph(pred_activated, pred_label, label, self.min_area, num_classes)

        multiclass_info_array = np.zeros((num_classes, 100, 3))  # 存储各种类别情况下的信息，一个三维矩阵

        for conf_i in [round(i * 0.01, 2) for i in range(0, 100)]:
            tp_nums, fp_nums = self.analyse_graph(graph, conf_i, self.iou_threshold, self.min_area, num_classes)
            pred_nums = tp_nums + fp_nums  # 统计检出区域的总数
            for subclass in range(num_classes):
                multiclass_info_array[subclass, int(conf_i * 100), 0] = int(gt_nums[subclass])
                multiclass_info_array[subclass, int(conf_i * 100), 1] = int(pred_nums[subclass])
                multiclass_info_array[subclass, int(conf_i * 100), 2] = int(tp_nums[subclass])
        return multiclass_info_array

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data and data_samples.

        The processed results should be stored in ``self.results``, which will
        be used to compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of outputs from the model.
        """
        for data_sample in data_samples:  # data_sample是单个的样本,在一个batch里面
            pred_logit = data_sample['seg_logits']['data']  # c, h, w
            pred_label = data_sample['pred_sem_seg']['data'].squeeze()  # h, w
            label = data_sample['gt_sem_seg']['data'].squeeze().to(pred_label)  # h, w
            self.results.append(self.overkill_and_escape(pred_logit, pred_label, label))
            if self.save_name is not None:
                # 只有在指定了保存文件名字的时候才会触发这个多置信度指标计算过程
                # 计算混淆矩阵, by lck
                multiclass_info_array = self.compute_region_paring(pred_logit, pred_label, label)  # num_classes, 100, 3
                if self.overall_info_array is None:
                    self.overall_info_array = multiclass_info_array
                else:
                    self.overall_info_array += multiclass_info_array

    def compute_metrics(self, results: list) -> Dict[str, float]:
        """Compute the metrics from processed results.
        Added functions:
            1. every classes' metrics into logger
            2. class 'placeholder' clean up.

        Args:
            results (list): The processed results of each batch.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
                the metrics, and the values are corresponding results. The key
                mainly includes aAcc, mIoU, mAcc, mDice, mFscore, mPrecision,
                mRecall.
        """
        if self.save_name is not None:
            # 在之前的基础上额外增加基于overall_info_array(num_classes, 100, 3)的信息处理与保存, gt pred tp
            recall_array = self.overall_info_array[:, :, 2] / self.overall_info_array[:, :, 0]  # [numclass, 100]
            precision_array = self.overall_info_array[:, :, 2] / self.overall_info_array[:, :, 1]  # [numclass, 100]
            f1_score_array = 2 * recall_array * precision_array / (precision_array + recall_array)  # [numclass, 100]

            # 暂时先只存全局的情况，三维矩阵没法存储
            result_array = np.vstack([np.array([round(i / 100., 2) for i in range(100)]),
                                      np.round(f1_score_array[-1], 3),
                                      np.round(recall_array[-1], 3),
                                      np.round(1 - precision_array[-1], 3)])  # 4, 100  第一行是置信度

            def save_xls(result_array, savename="list2Excel.xls"):
                import xlwt
                workbook = xlwt.Workbook()
                sheet = workbook.add_sheet("Sheet")

                for i in range(len(result_array)):
                    for j in range(len(result_array[i])):
                        sheet.write(i, j, result_array[i][j])

                workbook.save(savename)

            save_xls(result_array, self.save_name)

        logger: MMLogger = MMLogger.get_current_instance()

        # convert list of tuples to tuple of lists, e.g.
        # [(A_1, B_1, C_1, D_1), ...,  (A_n, B_n, C_n, D_n)] to
        # ([A_1, ..., A_n], ..., [D_1, ..., D_n])
        results = tuple(zip(*results))
        if len(results) != 4:
            return {}

        total_tp_region = sum(results[0])
        total_fp_region = sum(results[1])
        total_gt_region = sum(results[2])
        total_pred_region = sum(results[3])

        ret_metrics = OrderedDict()
        recall_rate = total_tp_region / total_gt_region
        precision_rate = 1 - total_fp_region / total_pred_region
        f_score = 2 * precision_rate * recall_rate / (precision_rate + recall_rate)

        ret_metrics['综合指标'] = f_score
        ret_metrics['区域检出率'] = recall_rate
        ret_metrics['区域过检率'] = 1 - precision_rate

        class_names = self.dataset_meta['classes'][1:]  # 不计算背景
        if not isinstance(self.dataset_meta['classes'], list):
            class_names = list(class_names)
        class_names.append('全局')
        metrics = dict()

        ret_metrics_class = OrderedDict({
            ret_metric: np.round(ret_metric_value * 100, 2)
            for ret_metric, ret_metric_value in ret_metrics.items()
        })

        # Add classes acc, iou into metrics [add by ldf]
        for key, val in ret_metrics_class.items():
            if key == '综合指标':  # 不想返回综合指标这个，这个东西打印显示就行
                continue
            for i, class_name in enumerate(class_names):
                metrics[key + '_' + class_name] = str(round(float(val[i]), 2))

        ret_metrics_class.update({'Class': class_names})
        ret_metrics_class.move_to_end('Class', last=False)

        class_table_data = PrettyTable()
        for key, val in ret_metrics_class.items():
            if '区域过检率' in key:  # 如果是precision\Recall，需要显示具体的区域数量
                val = list(map(str, val))  # 把数据都变成list(str)
                # 这个逻辑不想讲，反正结果是对的
                regions = ["(" + i + "/" + j + ")" for i, j in
                           zip(list(map(str, total_fp_region)),
                               list(map(str, total_pred_region)))]
                class_table_data.add_column(key, [i + ' ' + j for i, j in zip(val, regions)])
            elif '区域检出率' in key:
                val = list(map(str, val))  # 把数据都变成list(str)
                # 这个逻辑不想讲，反正结果是对的
                regions = ["(" + i + "/" + j + ")" for i, j in
                           zip(list(map(str, total_tp_region)),
                               list(map(str, total_gt_region)))]
                class_table_data.add_column(key, [i + ' ' + j for i, j in zip(val, regions)])
            else:
                class_table_data.add_column(key, val)

        print_log('per class results:', logger)
        print_log('\n' + class_table_data.get_string(), logger=logger)

        return metrics

    @staticmethod
    def build_graph(pred_activated: np.array,
                    pred_map: np.array,
                    label_map: np.array,
                    min_area: int,
                    num_classes: int) -> (nx.Graph, List, List):
        """根据给定的预测结果，构建无向图，该无向图可以包含类别错误但是面积上重合的目标.
        Args:
            pred_activated (np.array): 预测得到的置信度图
            pred_map (np.array): 预测结果图
            label_map (np.array): 标注图
            min_area (int): 最小面积
            num_classes (int): 类别数量
        Returns: (nx.Graph, List, List)
            graph: 无向图
            gt_nums: 待检测的缺陷数量，是个list，长度等于类别数量,最后一个是不管类别时候的数量
        """
        # 进来的，是个多类的 map 和 label都存在多类的可能性
        # 开始利用检测图、GT图和交集图，构建不同连通域之间关系的无向图，带有类别信息
        graph = nx.Graph()
        label_node_list = []  # GT 的不是双重list是因为不是按照类别分配index而是一起算的，索引的时候只能根据index不然太复杂
        gt_nums = [0] * num_classes  # -1号位准备存不管类别的情况下的区域数量
        pred_node_lists = [[] for _ in range(num_classes - 1)]  # -1 是从根本上不想计算和背景相关的东西
        pred_nums = [0] * num_classes

        # 先把所有的GT放进去，之所以先建立GT而不是和Pred一起建立，是因为可能存在类别不同但是交叠的情况
        label_map_single = morphology.remove_small_objects(label_map != 0, min_size=min_area).astype(np.uint8)
        label_nums, label_compnts, label_stats, _ = cv2.connectedComponentsWithStats(label_map_single, connectivity=8)  # noqa
        gt_nums[-1] = label_nums - 1
        for label_index in range(1, label_nums):
            label_class = (label_map * (label_compnts == label_index)).max()
            node_name = f'g{label_index}_{label_class}_{label_stats[label_index][-1]}'
            graph.add_node(node_name)  # g 结构是 gid_class_area
            label_node_list.append(node_name)  # noqa
            gt_nums[label_class - 1] += 1

        # 因为pred是可能存在完全粘连但是被分开判成两个类别的情况，因此pred_map不能先单类再读取类别这样搞，只能遍历类别这样
        # 同理相交图也是一样，所以也得分类别处理，再往图里添加，同时，建立起来与gt的连接关系
        for class_i in range(1, num_classes):
            # 每一张图中对每一个类别单独计算指标
            pred_map_i = morphology.remove_small_objects(pred_map == class_i, min_size=min_area).astype(np.uint8)

            inter_map_i = (pred_map_i * label_map_single).astype(np.uint8)

            pred_nums_i, pred_compnts_i, pred_stats_i, _ = cv2.connectedComponentsWithStats(pred_map_i)  # noqa
            inter_nums_i, inter_compnts_i, inter_stats_i, _ = cv2.connectedComponentsWithStats(inter_map_i)  # noqa
            pred_nums[class_i - 1] = pred_nums_i - 1
            pred_nums[-1] += pred_nums_i - 1
            for pred_index in range(1, pred_nums_i):  # 从1开始是因为0是背景
                score = (pred_activated * (pred_compnts_i == pred_index)).sum()
                node_name = f'p{pred_index}_{class_i}_{pred_stats_i[pred_index][-1]}_{score}'
                graph.add_node(node_name)  # p 的结构是pid_class_area_置信度总值
                pred_node_lists[class_i - 1].append(node_name)
            for inter_index in range(1, inter_nums_i):
                # 先存储这个inter区域的信息
                node_name = f'i{inter_index}_{class_i}_{inter_stats_i[inter_index][-1]}'
                graph.add_node(node_name)  # i 结构也是 iid_class_area

                # 然后得到这个inter区域对于的pred和gt的index
                inter_region = (inter_compnts_i == inter_index)  # 找出这个index对应的交集区域
                pred_index = (pred_compnts_i * inter_region).max()  # 根据这个交集区域，找出对应pred的index
                label_index = (label_compnts * inter_region).max()  # 以及对应的gt的index

                pred_node = pred_node_lists[class_i - 1][pred_index - 1]  # -1 是因为index是从1开始算的
                label_node = label_node_list[label_index - 1]  # 因为gt的index是全局的，没有类别这一层

                graph.add_edge(node_name, pred_node)  # 由于只会有1个GT和1个pred与这个交集区域相关，所以只需要添加这两个边就行
                graph.add_edge(node_name, label_node)

        # 至此完成图的创建，之所以要和后边的分析过程分开，是因为图建立完之后可能会使用不同的置信度阈值进行分析
        # 不返回pred的信息，因为有可能一个GT有俩Pred相交，或者俩gt有一个pred相交，pred自适应根据gt数量变化 advised from ldf
        return graph, np.array(gt_nums)

    @staticmethod
    def analyse_graph(graph, score_threshold, iou_threshold, min_area, num_classes):
        # 注意，这里的TP区域还是按照标注的区域数量来算的，两个GT如果靠近导致Pred粘连，也是两个一起检出或者不检出
        tp_nums = [0] * num_classes  # 这个长度和num_class一样，是因为-1号准备存全部当单类时候的值
        # 这个就单纯按照pred的区域来算就行
        fp_nums = [0] * num_classes  # 同上

        # 得到图之后，分析每一个连通子图，每个连通子图里边，多个GT和多个pred都联合起来算IOU和置信度，决定这批GT是否检出
        # 这个子图比较恶心，可能有不同类别的Pred和gt相交
        for sub_graph in nx.connected_components(graph):
            # 下面这些都是一起计算检出与否,需要遍历来统计几项内容，
            # GT总面积，GT区域数量，Pred总面积，Pred区域数量，交集总面积，置信度总值，这些都得是list,因为要存不同类别的情况。。。
            gt_area, gt_related_nums, pred_area, pred_related_nums, inter_area, score_sum = \
                [[0] * num_classes for _ in range(6)]
            for node in sub_graph:
                node_type = node[0]  #
                node_info = node.split('_')
                node_class = int(node_info[1])
                node_area = int(node_info[2])
                if node_type == 'i':
                    inter_area[node_class - 1] += node_area
                    inter_area[-1] += node_area  # -1存储的是类别无关的信息
                elif node_type == 'g':
                    gt_area[node_class - 1] += node_area
                    gt_related_nums[node_class - 1] += 1
                    gt_area[-1] += node_area
                    gt_related_nums[-1] += 1
                else:
                    score = float(node_info[-1])
                    pred_area[node_class - 1] += node_area
                    pred_related_nums[node_class - 1] += 1
                    score_sum[node_class - 1] += score
                    pred_area[-1] += node_area
                    pred_related_nums[-1] += 1
                    score_sum[-1] += score
            # 如果这个连通图存在交集区域，那么pred区域按照gt区域数量来算，不然还是实际的pred区域数量
            pred_related_nums = [gt_related_nums[i] if inter_area[i] != 0 else pred_related_nums[i]
                                 for i in range(num_classes)]

            for index in range(num_classes):  # 不想起名叫subclass 因为最后一个位置不是类别的意思
                # 过完这个子图的所有点了，分类分析，就三种情况，1.只有GT 2.只有pred 3.有交集
                if pred_related_nums[index] == 0:
                    continue  # (纯GT的情况这个不用单独处理，只要统计哪些GT检测出来就行)
                # 剩下两种有pred的情况
                # 如果只有Pred没有GT，按照过检来分析
                if gt_related_nums[index] == 0:
                    if pred_area[index] > min_area and score_sum[index] / pred_area[index] > score_threshold:
                        fp_nums[index] += pred_related_nums[index]
                else:  # 这个按照置信度+IOU判定是否检出
                    # 因为有pred有GT必定有inter区域
                    iou = inter_area[index] / (pred_area[index] + gt_area[index] - inter_area[index])
                    if score_sum[index] / pred_area[index] > score_threshold and iou > iou_threshold:
                        tp_nums[index] += gt_related_nums[index]  # 把这部分检出的缺陷数量加到TP里面去
        # 存的是各个类别以及不算类别情况下的信息
        return np.array(tp_nums), np.array(fp_nums)

    def overkill_and_escape(self,
                            pred_logit: torch.tensor,  # (c, h, w)
                            pred_label: torch.tensor,
                            label: torch.tensor,
                            ):
        """Calculate Intersection and Union.
        pred_logit 已经被激活了，不用再处理
        num_classes 包含了背景类别

        返回的，是每个类别的检出区域数量、过检区域数量、GT区域数量和预测区域数量
        """

        pred_activated = pred_logit.max(dim=0, keepdim=True)[0].squeeze()
        pred_activated = pred_activated.cpu().numpy()  # (h, w)
        mask = (label != self.ignore_index)
        pred_label = (pred_label * mask).cpu().numpy()
        label = (label * mask).cpu().numpy()
        num_classes = len(self.dataset_meta['classes'])

        # 得到无向图，和对应的GT还有检测区域的数量
        graph, gt_nums = self.build_graph(pred_activated, pred_label, label, self.min_area, num_classes)

        tp_nums, fp_nums = self.analyse_graph(graph, self.score_threshold, self.iou_threshold, self.min_area, num_classes)

        return tp_nums, fp_nums, gt_nums, tp_nums + fp_nums


@METRICS.register_module()
class VisaionMetricHttp(VisaionMetric):
    """detail ref DLPMetric
    """

    def __init__(self, iou_threshold=0.3, no_argmax=False, output_dir=None, **kwargs) -> None:
        super(VisaionMetricHttp, self).__init__(**kwargs)
        # 设置输出目录
        if os.environ.get("VISAION_WORK_DIR", None):
            self.output_dir = os.environ.get("VISAION_WORK_DIR", None)
        else:
            self.output_dir = output_dir
        os.makedirs(self.output_dir, exist_ok=True)
        self.output_dir_region_paring = osp.join(self.output_dir, "region_paring")
        os.makedirs(self.output_dir_region_paring, exist_ok=True)
        self.iou_threshold = iou_threshold
        self.no_argmax = no_argmax

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        """Process one batch of data and data_samples.

        The processed results should be stored in ``self.results``, which will
        be used to compute the metrics when all batches have been processed.

        Args:
            data_batch (dict): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of outputs from the model.
        """
        for data_sample in data_samples:  # data_sample是单个的样本
            pred_logit = data_sample['seg_logits']['data']  # c, h, w
            pred_label = data_sample['pred_sem_seg']['data'].squeeze()  # h, w
            label = data_sample['gt_sem_seg']['data'].squeeze().to(pred_label)  # h, w

            # 获取样本ID
            # sample_id = data_samples[0].get('sample_global_id', 'repeat')
            sample_id = Path(data_sample['img_path']).stem    # 使用图片名作为sample_id
            # sample_id = data_sample['img_id']  # 使用图片id作为sample_id

            # 存储Origin预测结果
            seg_logits_ = pred_logit.detach().cpu().numpy()
            dst_path = os.path.join(self.output_dir, "pred_origin", str(sample_id)+".npy")
            os.makedirs(os.path.dirname(dst_path), exist_ok=True)
            np.save(dst_path, seg_logits_)

            # 存储Argmax预测结果
            pred_sem_seg = pred_logit.detach().cpu()
            if self.no_argmax:
                _label = label.detach().cpu().numpy().astype(np.uint8)
                _conf = _label * pred_logit.squeeze().detach().cpu().numpy() * 255
            else:
                _label = np.argmax(pred_sem_seg.numpy(), axis=0).astype(np.uint8)
                _conf = (np.max(pred_sem_seg.numpy(), axis=0) * 255).astype(np.uint8)
                _conf = np.where(_label != 0, 1, 0) * _conf

            # 标记前景区域
            labeled_array, num_features = sclabel(_conf > 0)

            # 计算每个区域的均值
            means = scmean(_conf, labeled_array, index=np.arange(1, num_features + 1))

            # 创建一个与输入图像大小相同的输出图像
            output_image = np.zeros_like(_conf)

            # 将均值填充到相应的区域
            for i in range(1, num_features + 1):
                output_image[labeled_array == i] = means[i - 1]

            mask = np.stack([_label, output_image, _conf], axis=0)
            mask = mask.transpose(1, 2, 0)
            dst_path2 = os.path.join(self.output_dir, "pred_origin", str(sample_id) + ".png")
            os.makedirs(os.path.dirname(dst_path2), exist_ok=True)
            cv2.imwrite(dst_path2, mask)

            # 如果未标注的图片 继续下一张
            if torch.all(label == 255):
                continue
            self.results.append(self.overkill_and_escape(pred_logit, pred_label, label))

            # 计算混淆矩阵, by lck
            multiclass_info_list = self.compute_region_paring(pred_logit, pred_label, label)
            self.save_json(multiclass_info_list, osp.join(self.output_dir_region_paring, str(sample_id) + ".json"))

    def compute_metrics(self, results: list) -> Dict[str, float]:
        """Compute the metrics from processed results.
        Added functions:
            1. every classes' metrics into logger
            2. class 'placeholder' clean up.

        Args:
            results (list): The processed results of each batch.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
                the metrics, and the values are corresponding results. The key
                mainly includes aAcc, mIoU, mAcc, mDice, mFscore, mPrecision,
                mRecall.
        """
        logger: MMLogger = MMLogger.get_current_instance()

        # convert list of tuples to tuple of lists, e.g.
        # [(A_1, B_1, C_1, D_1), ...,  (A_n, B_n, C_n, D_n)] to
        # ([A_1, ..., A_n], ..., [D_1, ..., D_n])
        results = tuple(zip(*results))
        if len(results) != 4:
            return {}

        total_tp_region = sum(results[0])
        total_fp_region = sum(results[1])
        total_gt_region = sum(results[2])
        total_pred_region = sum(results[3])

        ret_metrics = OrderedDict()
        recall_rate = total_tp_region / total_gt_region
        precision_rate = 1 - total_fp_region / total_pred_region
        f_score = 2 * precision_rate * recall_rate / (precision_rate + recall_rate)

        ret_metrics['综合指标'] = f_score
        ret_metrics['区域检出率'] = recall_rate
        ret_metrics['区域过检率'] = 1 - precision_rate

        class_names = self.dataset_meta['classes'][1:]  # 不计算背景
        if not isinstance(self.dataset_meta['classes'], list):
            class_names = list(class_names)
        class_names.append('全局')
        metrics = dict()

        ret_metrics_class = OrderedDict({
            ret_metric: np.round(ret_metric_value * 100, 2)
            for ret_metric, ret_metric_value in ret_metrics.items()
        })

        # Add classes acc, iou into metrics [add by ldf]
        for key, val in ret_metrics_class.items():
            if key == '综合指标':  # 不想返回综合指标这个，这个东西打印显示就行
                continue
            for i, class_name in enumerate(class_names):
                metrics[key + '_' + class_name] = str(round(float(val[i]), 2))

        ret_metrics_class.update({'Class': class_names})
        ret_metrics_class.move_to_end('Class', last=False)

        class_table_data = PrettyTable()
        for key, val in ret_metrics_class.items():
            if '区域过检率' in key:  # 如果是precision\Recall，需要显示具体的区域数量
                val = list(map(str, val))  # 把数据都变成list(str)
                # 这个逻辑不想讲，反正结果是对的
                regions = ["(" + i + "/" + j + ")" for i, j in
                           zip(list(map(str, total_fp_region)),
                               list(map(str, total_pred_region)))]
                class_table_data.add_column(key, [i + ' ' + j for i, j in zip(val, regions)])
            elif '区域检出率' in key:
                val = list(map(str, val))  # 把数据都变成list(str)
                # 这个逻辑不想讲，反正结果是对的
                regions = ["(" + i + "/" + j + ")" for i, j in
                           zip(list(map(str, total_tp_region)),
                               list(map(str, total_gt_region)))]
                class_table_data.add_column(key, [i + ' ' + j for i, j in zip(val, regions)])
            else:
                class_table_data.add_column(key, val)

        print_log('per class results:', logger)
        print_log('\n' + class_table_data.get_string(), logger=logger)
        self.save_json({
            "iou_threshold": self.iou_threshold,
            "min_area": self.min_area
        }, osp.join(self.output_dir_region_paring, "kwargs.json"))

        return metrics

    def compute_region_paring(self,
                              pred_logit: torch.tensor,
                              pred_label: torch.tensor,
                              label: torch.tensor
                              ):
        """Calculate Confusion Matrix.
            Args:
                pred_logit (torch.tensor): Prediction origin after activation. The shape is (numClass, H, W).
                pred_label (torch.tensor): Prediction results
                label (torch.tensor): Ground truth segmentation map. The shape is (H, W).
            Returns:
                multiclass_info_list (list)
        """

        pred_activated = pred_logit.max(dim=0, keepdim=True)[0].squeeze()
        pred_activated = pred_activated.cpu().numpy()  # (h, w)
        mask = (label != self.ignore_index)
        pred_label = (pred_label * mask).cpu().numpy()
        label = (label * mask).cpu().numpy()
        num_classes = pred_logit.shape[0]

        # 得到无向图，和对应GT的数量
        graph, gt_nums = self.build_graph(pred_activated, pred_label, label, self.min_area, num_classes)

        multiclass_info_list = []  # 存储各种类别情况下的信息 最后一个存的是不分类别情况下的值

        for conf_i in [round(i * 0.01, 2) for i in range(0, 100)]:
            tp_nums, fp_nums = self.analyse_graph(graph, conf_i, self.iou_threshold, self.min_area, num_classes)
            pred_nums = tp_nums + fp_nums  # 统计检出区域的总数
            subclass_info = []
            for subclass in range(num_classes):
                # 不分类的subclass按-1存
                subclass_info.append(
                    {'DefectId': subclass if subclass < num_classes-1 else -1,
                     'gt_nums': int(gt_nums[subclass]),  # 缺陷区域
                     'pred_nums': int(pred_nums[subclass]),  # 全部检测区域
                     'tp_nums': int(tp_nums[subclass])}
                )  # 正确检出区域
            multiclass_info_list.append(subclass_info)
        return multiclass_info_list

    @staticmethod
    def save_json(json_dict: Union[Dict, List] = None, save_path: str = None) -> None:
        """
        保存dict为json文件
        :param json_dict:   字典
        :param save_path:   保存路径
        :return:
        """
        with open(save_path, 'w', encoding="utf-8") as fp:
            json.dump(json_dict, fp, ensure_ascii=False, indent=4)


@METRICS.register_module()
class TransStabilityMetric(BaseMetric):
    def __init__(self,
                 ignore_index: int = 255,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None,
                 ):
        super().__init__(collect_device=collect_device, prefix=prefix)
        self.ignore_index = ignore_index

    def process(self, data_batch: dict, data_samples: Sequence[dict]) -> None:
        # data_samples: area_list, prob_list
        if len(data_samples) == 0:
            return  # 负样本
        area_list, prob_list = data_samples
        area_mean, area_std = area_list.mean(), area_list.std()
        prob_mean, prob_std = prob_list.mean(), prob_list.std()

        area_cov = area_std / (area_mean + 1e-8)  # cov:变异系数，也就是波动性的大小
        prob_cov = prob_std / (prob_mean + 1e-8)  # cov:变异系数，也就是波动性的大小

        self.results.append([area_mean, area_std, area_cov, prob_mean, prob_std, prob_cov])

    def compute_metrics(self, results: list) -> Dict[str, float]:
        logger: MMLogger = MMLogger.get_current_instance()
        res = tuple(zip(*results))
        _, m_area_std, m_area_cov, m_prob_mean, m_prob_std, m_prob_cov = [sum(res[i]) / len(res[i]) for i in range(6)]

        ret_metrics = OrderedDict()
        ret_metrics['平均面积标准差'] = m_area_std
        ret_metrics['平均面积波动百分比'] = m_area_cov * 100.
        ret_metrics['平均置信度均值'] = m_prob_mean
        ret_metrics['平均置信度标准差'] = m_prob_std
        ret_metrics['平均置信度波动百分比'] = m_prob_cov * 100.

        class_table_data = PrettyTable()
        for key, val in ret_metrics.items():
            class_table_data.add_column(key, [str(np.round(val, 3))])
        print_log('\n' + class_table_data.get_string(), logger=logger)
        metrics = dict()
        for key, val in ret_metrics.items():
            metrics[key] = round(val, 3)
        return metrics



import os
import numpy as np
from PIL import Image
import torch
import torchvision
from visaionlibrary.utils.yoloe import YOLOE
from visaionlibrary.utils.yoloe.models.yolo.yoloe.predict_vp import YOLOEVPSegPredictor


class YoloE:
    def __init__(self, model_path="yoloe-v8l-seg.pt", device="cuda:0"):
        VISAION_DIR = os.getenv("VISAION_DIR")
        if VISAION_DIR is None:
            raise ValueError("VISAION_DIR is not set")
        self.model_path = os.path.join(VISAION_DIR, "weights", model_path)
        self.device = device
        self.model = YOLOE(self.model_path)
        self.model.to(self.device)

    def _apply_nms(self, class_names, confidence, bboxes, masks, iou_threshold=0.5):
        """
        Apply Non-Maximum Suppression (NMS) to filter overlapping bounding boxes.
        
        Args:
            class_names: list[str] - List of class names for each detection
            confidence: np.ndarray - Confidence scores, shape (N,)
            bboxes: np.ndarray - Bounding boxes in xyxy format, shape (N, 4)
            masks: np.ndarray or None - Masks data, shape (N, H, W) or None
            iou_threshold: float - IoU threshold for NMS (default: 0.6)
            
        Returns:
            tuple: (filtered_class_names, filtered_confidence, filtered_bboxes, filtered_masks)
        """
        if len(bboxes) == 0:
            return class_names, confidence, bboxes, masks
        
        # Convert to torch tensors
        bboxes_tensor = torch.from_numpy(bboxes).float()
        confidence_tensor = torch.from_numpy(confidence).float()
        
        # Group by class and apply NMS per class
        unique_classes = list(set(class_names))
        keep_indices = []
        
        for cls in unique_classes:
            # Get indices for this class
            cls_indices = [i for i, name in enumerate(class_names) if name == cls]
            
            if len(cls_indices) == 0:
                continue
            
            # Get boxes and scores for this class
            cls_boxes = bboxes_tensor[cls_indices]
            cls_scores = confidence_tensor[cls_indices]
            
            # Sort by confidence (descending)
            sorted_idx = torch.argsort(cls_scores, descending=True)
            cls_boxes = cls_boxes[sorted_idx]
            cls_scores = cls_scores[sorted_idx]
            cls_indices_sorted = [cls_indices[i] for i in sorted_idx]
            
            # Apply NMS
            if len(cls_boxes) > 0:
                keep = torchvision.ops.nms(cls_boxes, cls_scores, iou_threshold)
                # Convert keep indices back to original indices
                # keep is a tensor of indices relative to sorted cls_boxes
                # cls_indices_sorted contains the original indices in sorted order
                keep_numpy = keep.cpu().numpy()
                keep_indices.extend([cls_indices_sorted[i] for i in keep_numpy])
        
        # Sort keep_indices to maintain order (by original index)
        keep_indices = sorted(keep_indices)
        
        # Filter all arrays
        filtered_class_names = [class_names[i] for i in keep_indices]
        filtered_confidence = confidence[keep_indices]
        filtered_bboxes = bboxes[keep_indices]
        
        if masks is not None:
            filtered_masks = masks[keep_indices]
        else:
            filtered_masks = None
        
        return filtered_class_names, filtered_confidence, filtered_bboxes, filtered_masks

    def get_visual_pe(self, prompt_images:list[Image.Image], prompt_boxes:dict[str, list[np.ndarray]], prompt_names:list[str], **kwargs):
        """
        Get visual prompt embeddings from prompt images.
        Args:
            prompt_images: list[Image.Image]
            prompt_boxes: dict[str, list[np.ndarray]]
            prompt_names: list[str]
            **kwargs: dict
        Returns:
            prompt_names: list[list[str]] - flattened list of all class names
            visual_pe: list[np.ndarray] - list of visual prompt embeddings
        """
        assert "bboxes" in prompt_boxes and "cls" in prompt_boxes, f"prompt_boxes must contain 'bboxes' and 'cls'"
        assert len(prompt_images) == len(prompt_boxes["bboxes"]) == len(prompt_boxes["cls"]) == len(prompt_names), f"the length of prompt_image, prompt_boxes['bboxes'], prompt_boxes['cls'], and prompt_names must be the same"

        visual_pe_list = []
        for prompt_image, prompt_box, prompt_cls, prompt_name in zip(prompt_images, prompt_boxes["bboxes"], prompt_boxes["cls"], prompt_names):
            visual_prompt = dict(
                bboxes=prompt_box,
                cls=prompt_cls,
            )

            # Reset predictor before getting vpe to ensure YOLOEVPSegPredictor is used
            self.model.predictor = None
            self.model.predict(prompt_image, prompts=visual_prompt, predictor=YOLOEVPSegPredictor, return_vpe=True, **kwargs)
            # Get vpe from predictor immediately
            vpe = self.model.predictor.vpe
            vpe_np = vpe.detach().cpu().numpy()
            visual_pe_list.append(vpe_np)
        
        return prompt_names, visual_pe_list

    def visual(self, images:list[Image.Image], prompt_names:list[list[str]], visual_pe:list[np.ndarray], **kwargs):
        """
        Predict the visual prompt of the images.
        Args:
            images: list[Image.Image]
            prompt_names: list[list[str]]
            visual_pe: list[np.ndarray]
            **kwargs: dict
        Returns:
            list[Image.Image]
            list[str]
            list[float]
            list[list[float]]
            list[list[list[float]]]
        """
        assert len(prompt_names) == len(visual_pe), f"the length of prompt_names and visual_pe must be the same"

        # Get batch_size from kwargs, default to 8 to avoid OOM
        batch_size = kwargs.pop('batch_size', 8)
        
        for prompt_name, vpe_np in zip(prompt_names, visual_pe):
            vpe = torch.from_numpy(vpe_np).to(self.device)
            self.model.set_classes(prompt_name, vpe)
            self.model.predictor = None  # Reset predictor to apply new class count
            # Disable fuse when class count changes to avoid shape mismatch
            
            # Process images in batches to avoid OOM
            total_images = len(images)
            for batch_start in range(0, total_images, batch_size):
                batch_end = min(batch_start + batch_size, total_images)
                batch_images = images[batch_start:batch_end]
                
                # Process batch
                results = self.model.predict(batch_images, save=False, fuse=False, **kwargs)
                
                # Yield results with correct image_index
                for batch_idx, result in enumerate(results):
                    image_index = batch_start + batch_idx
                    classes_index = result.boxes.cls.detach().cpu().numpy()
                    class_names = [prompt_name[int(class_index)] for class_index in classes_index]
                    confidence = result.boxes.conf.detach().cpu().numpy()
                    bboxes = result.boxes.xyxy.detach().cpu().numpy()
                    if result.masks is not None:
                        masks = result.masks.cpu().numpy()
                    else:
                        masks = None
                
                # Apply NMS to filter overlapping boxes (iou_threshold=0.5)
                class_names, confidence, bboxes, masks = self._apply_nms(
                    class_names, confidence, bboxes, masks, iou_threshold=0.5
                )
                
                yield image_index, class_names, confidence, bboxes, masks
                
                # Clear GPU cache after each batch to free memory
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()

    def get_text_pe(self, prompt:list[str], **kwargs):
        """
        Predict the text prompt of the images.
        Args:
            prompt: list[str]
            **kwargs: dict
        Returns:
            np.ndarray
        """
        assert self.model is not None, "Model is not initialized"
        text_pe = self.model.get_text_pe(prompt)
        text_pe = text_pe.detach().cpu().numpy()
        return prompt, text_pe

    def text(self, images:list[Image.Image], prompt:list[str], text_pe:np.ndarray, **kwargs):
        """
        Predict the visual prompt of the images.
        Args:
            images: list[Image.Image]
            prompt: list[str]
            **kwargs: dict
        Returns:
            list[Image.Image]
            list[str]
            list[float]
            list[list[float]]
            list[list[list[float]]]
        """
        text_pe = torch.from_numpy(text_pe).to(self.device)
        assert text_pe.ndim == 3, "text_pe must be a 3D tensor"
        self.model.set_classes(prompt, text_pe)
        self.model.predictor = None  # Reset predictor to apply new class count
        # Disable fuse when class count changes to avoid shape mismatch
        
        # Get batch_size from kwargs, default to 8 to avoid OOM
        batch_size = kwargs.pop('batch_size', 8)
        
        # Process images in batches to avoid OOM
        total_images = len(images)
        for batch_start in range(0, total_images, batch_size):
            batch_end = min(batch_start + batch_size, total_images)
            batch_images = images[batch_start:batch_end]
            
            # Process batch
            results = self.model.predict(batch_images, verbose=False, fuse=False, **kwargs)
            
            # Yield results with correct image_index
            for batch_idx, result in enumerate(results):
                image_index = batch_start + batch_idx
                classes_index = result.boxes.cls.detach().cpu().numpy()
                class_names = [prompt[int(class_index)] for class_index in classes_index]
                confidence = result.boxes.conf.detach().cpu().numpy()
                bboxes = result.boxes.xyxy.detach().cpu().numpy()
                if result.masks is not None:
                    masks = result.masks.cpu().numpy()
                else:
                    masks = None
                
                # Apply NMS to filter overlapping boxes (iou_threshold=0.5)
                class_names, confidence, bboxes, masks = self._apply_nms(
                    class_names, confidence, bboxes, masks, iou_threshold=0.5
                )
                
                yield image_index, class_names, confidence, bboxes, masks
            
            # Clear GPU cache after each batch to free memory
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

    def predict(self, images:list[Image.Image], prompt_names:list[list[str]], pe:list[np.ndarray], **kwargs):
        """
        Predict the mixed prompt of the images.
        Args:
            images: list[Image.Image]
            prompt_names: list[list[str]]
            pe: list[np.ndarray]
            **kwargs: dict
        Returns:
            list[Image.Image]
            list[str]
            list[float]
            list[list[float]]
            list[list[list[float]]]
        """
        assert len(prompt_names) == pe.shape[1], f"the length of prompt_names and pe must be the same"

        # Get batch_size from kwargs, default to 8 to avoid OOM
        batch_size = kwargs.pop('batch_size', 8)

        pe = torch.from_numpy(pe).to(self.device)
        self.model.set_classes(prompt_names, pe)
        self.model.predictor = None  # Reset predictor to apply new class count
        # Disable fuse when class count changes to avoid shape mismatch

        # Process images in batches to avoid OOM
        total_images = len(images)
        for batch_start in range(0, total_images, batch_size):
            batch_end = min(batch_start + batch_size, total_images)
            batch_images = images[batch_start:batch_end]
            
            # Process batch
            results = self.model.predict(batch_images, save=False, fuse=False, **kwargs)
            
            # Yield results with correct image_index
            for batch_idx, result in enumerate(results):
                image_index = batch_start + batch_idx
                classes_index = result.boxes.cls.detach().cpu().numpy()
                class_names = [prompt_names[int(class_index)] for class_index in classes_index]
                confidence = result.boxes.conf.detach().cpu().numpy()
                bboxes = result.boxes.xyxy.detach().cpu().numpy()
                if result.masks is not None:
                    masks = result.masks.cpu().numpy()
                else:
                    masks = None
                
                # Apply NMS to filter overlapping boxes (iou_threshold=0.5)
                class_names, confidence, bboxes, masks = self._apply_nms(
                    class_names, confidence, bboxes, masks, iou_threshold=0.5
                )
                
                yield image_index, class_names, confidence, bboxes, masks
            
            # Clear GPU cache after each batch to free memory
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
AILabel = YoloE


import contextlib
import os
import platform
import re
import threading
import time
import urllib
from pathlib import Path
from types import SimpleNamespace

import cv2
import numpy as np
import torch
import yaml


# Other Constants
FILE = Path(__file__).resolve()
ROOT = FILE.parents[1]  # YOLO
NUM_THREADS = min(8, max(1, os.cpu_count() - 1))  # number of YOLO multiprocessing threads
VERBOSE = str(os.getenv("YOLO_VERBOSE", True)).lower() == "true"  # global verbose mode
MACOS, LINUX, WINDOWS = (platform.system() == x for x in ["Darwin", "Linux", "Windows"])  # environment booleans
PYTHON_VERSION = platform.python_version()

# Settings and Environment Variables
torch.set_printoptions(linewidth=320, precision=4, profile="default")
np.set_printoptions(linewidth=320, formatter={"float_kind": "{:11.5g}".format})  # format short g, %precision=5
cv2.setNumThreads(0)  # prevent OpenCV from multithreading (incompatible with PyTorch DataLoader)
os.environ["NUMEXPR_MAX_THREADS"] = str(NUM_THREADS)  # NumExpr max threads
os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"  # for deterministic training to avoid CUDA warning
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # suppress verbose TF compiler warnings in Colab
os.environ["TORCH_CPP_LOG_LEVEL"] = "ERROR"  # suppress "NNPACK.cpp could not initialize NNPACK" warnings
os.environ["KINETO_LOG_LEVEL"] = "5"  # suppress verbose PyTorch profiler output when computing FLOPs



class SimpleClass:
    """
    A simple base class for creating objects with string representations of their attributes.

    This class provides a foundation for creating objects that can be easily printed or represented as strings,
    showing all their non-callable attributes. It's useful for debugging and introspection of object states.

    Methods:
        __str__: Returns a human-readable string representation of the object.
        __repr__: Returns a machine-readable string representation of the object.
        __getattr__: Provides a custom attribute access error message with helpful information.

    Examples:
        >>> class MyClass(SimpleClass):
        ...     def __init__(self):
        ...         self.x = 10
        ...         self.y = "hello"
        >>> obj = MyClass()
        >>> print(obj)
        __main__.MyClass object with attributes:

        x: 10
        y: 'hello'

    Notes:
        - This class is designed to be subclassed. It provides a convenient way to inspect object attributes.
        - The string representation includes the module and class name of the object.
        - Callable attributes and attributes starting with an underscore are excluded from the string representation.
    """

    def __str__(self):
        """Return a human-readable string representation of the object."""
        attr = []
        for a in dir(self):
            v = getattr(self, a)
            if not callable(v) and not a.startswith("_"):
                if isinstance(v, SimpleClass):
                    # Display only the module and class name for subclasses
                    s = f"{a}: {v.__module__}.{v.__class__.__name__} object"
                else:
                    s = f"{a}: {repr(v)}"
                attr.append(s)
        return f"{self.__module__}.{self.__class__.__name__} object with attributes:\n\n" + "\n".join(attr)

    def __repr__(self):
        """Return a machine-readable string representation of the object."""
        return self.__str__()

    def __getattr__(self, attr):
        """Custom attribute access error message with helpful information."""
        name = self.__class__.__name__
        raise AttributeError(f"'{name}' object has no attribute '{attr}'. See valid attributes below.\n{self.__doc__}")


class IterableSimpleNamespace(SimpleNamespace):
    """
    An iterable SimpleNamespace class that provides enhanced functionality for attribute access and iteration.

    This class extends the SimpleNamespace class with additional methods for iteration, string representation,
    and attribute access. It is designed to be used as a convenient container for storing and accessing
    configuration parameters.

    Methods:
        __iter__: Returns an iterator of key-value pairs from the namespace's attributes.
        __str__: Returns a human-readable string representation of the object.
        __getattr__: Provides a custom attribute access error message with helpful information.
        get: Retrieves the value of a specified key, or a default value if the key doesn't exist.

    Examples:
        >>> cfg = IterableSimpleNamespace(a=1, b=2, c=3)
        >>> for k, v in cfg:
        ...     print(f"{k}: {v}")
        a: 1
        b: 2
        c: 3
        >>> print(cfg)
        a=1
        b=2
        c=3
        >>> cfg.get("b")
        2
        >>> cfg.get("d", "default")
        'default'

    Notes:
        This class is particularly useful for storing configuration parameters in a more accessible
        and iterable format compared to a standard dictionary.
    """

    def __iter__(self):
        """Return an iterator of key-value pairs from the namespace's attributes."""
        return iter(vars(self).items())

    def __str__(self):
        """Return a human-readable string representation of the object."""
        return "\n".join(f"{k}={v}" for k, v in vars(self).items())

    def __getattr__(self, attr):
        """Custom attribute access error message with helpful information."""
        name = self.__class__.__name__
        raise AttributeError(
            f"""
            '{name}' object has no attribute '{attr}'. """
        )

    def get(self, key, default=None):
        """Return the value of the specified key if it exists; otherwise, return the default value."""
        return getattr(self, key, default)


def emojis(string=""):
    """Return platform-dependent emoji-safe version of string."""
    return string.encode().decode("ascii", "ignore") if WINDOWS else string


class ThreadingLocked:
    """
    A decorator class for ensuring thread-safe execution of a function or method. This class can be used as a decorator
    to make sure that if the decorated function is called from multiple threads, only one thread at a time will be able
    to execute the function.

    Attributes:
        lock (threading.Lock): A lock object used to manage access to the decorated function.

    Example:
        ```python
        from visaionlibrary.utils.yoloe.utils import ThreadingLocked

        @ThreadingLocked()
        def my_function():
            # Your code here
        ```
    """

    def __init__(self):
        """Initializes the decorator class for thread-safe execution of a function or method."""
        self.lock = threading.Lock()

    def __call__(self, f):
        """Run thread-safe execution of function or method."""
        from functools import wraps

        @wraps(f)
        def decorated(*args, **kwargs):
            """Applies thread-safety to the decorated function or method."""
            with self.lock:
                return f(*args, **kwargs)

        return decorated


def yaml_save(file="data.yaml", data=None, header=""):
    """
    Save YAML data to a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        data (dict): Data to save in YAML format.
        header (str, optional): YAML header to add.

    Returns:
        (None): Data is saved to the specified file.
    """
    if data is None:
        data = {}
    file = Path(file)
    if not file.parent.exists():
        # Create parent directories if they don't exist
        file.parent.mkdir(parents=True, exist_ok=True)

    # Convert Path objects to strings
    valid_types = int, float, str, bool, list, tuple, dict, type(None)
    for k, v in data.items():
        if not isinstance(v, valid_types):
            data[k] = str(v)

    # Dump data to file in YAML format
    with open(file, "w", errors="ignore", encoding="utf-8") as f:
        if header:
            f.write(header)
        yaml.safe_dump(data, f, sort_keys=False, allow_unicode=True)


def yaml_load(file="data.yaml", append_filename=False):
    """
    Load YAML data from a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        append_filename (bool): Add the YAML filename to the YAML dictionary. Default is False.

    Returns:
        (dict): YAML data and file name.
    """
    assert Path(file).suffix in {".yaml", ".yml"}, f"Attempting to load non-YAML file {file} with yaml_load()"
    with open(file, errors="ignore", encoding="utf-8") as f:
        s = f.read()  # string

        # Remove special characters
        if not s.isprintable():
            s = re.sub(r"[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\uD7FF\uE000-\uFFFD\U00010000-\U0010ffff]+", "", s)

        # Add YAML filename to dict and return
        data = yaml.safe_load(s) or {}  # always return a dict (yaml.safe_load() may return None for empty files)
        if append_filename:
            data["yaml_file"] = str(file)
        return data


# Default configuration
DEFAULT_CFG_DICT = {
    'task': 'detect', 'mode': 'train', 'model': None, 'data': None, 'epochs': 100, 'time': None, 
    'patience': 100, 'batch': 16, 'imgsz': 640, 'save': True, 'save_period': -1, 'cache': False, 
    'device': None, 'workers': 8, 'project': None, 'name': None, 'exist_ok': False, 'pretrained': True, 
    'optimizer': 'auto', 'verbose': True, 'seed': 0, 'deterministic': True, 'single_cls': False, 
    'rect': False, 'cos_lr': False, 'close_mosaic': 10, 'resume': False, 'amp': True, 'fraction': 1.0, 
    'profile': False, 'freeze': 'None', 'multi_scale': False, 'overlap_mask': False, 'mask_ratio': 4, 
    'dropout': 0.0, 'val': True, 'split': 'val', 'save_json': False, 'save_hybrid': False, 'conf': None, 
    'iou': 0.7, 'max_det': 300, 'half': False, 'dnn': False, 'plots': True, 'source': None, 
    'vid_stride': 1, 'stream_buffer': False, 'visualize': False, 'augment': False, 'agnostic_nms': False, 
    'classes': None, 'retina_masks': False, 'embed': None, 'show': False, 'save_frames': False, 'save_txt': False, 
    'save_conf': False, 'save_crop': False, 'show_labels': True, 'show_conf': True, 'show_boxes': True, 
    'line_width': None, 'format': 'torchscript', 'keras': False, 'optimize': False, 'int8': False, 
    'dynamic': False, 'simplify': True, 'opset': None, 'workspace': 'None', 'nms': False, 'lr0': 0.01, 'lrf': 0.01, 
    'momentum': 0.937, 'weight_decay': 0.0005, 'warmup_epochs': 3.0, 'warmup_momentum': 0.8, 'warmup_bias_lr': 0.1, 
    'box': 7.5, 'cls': 0.5, 'dfl': 1.5, 'pose': 12.0, 'kobj': 1.0, 'nbs': 64, 'hsv_h': 0.015, 'hsv_s': 0.7, 'hsv_v': 0.4, 
    'degrees': 0.0, 'translate': 0.1, 'scale': 0.5, 'shear': 0.0, 'perspective': 0.0, 'flipud': 0.0, 'fliplr': 0.5, 
    'bgr': 0.0, 'mosaic': 1.0, 'mixup': 0.0, 'copy_paste': 0.0, 'copy_paste_mode': 'flip', 'auto_augment': 'randaugment', 
    'erasing': 0.4, 'crop_fraction': 1.0, 'cfg': None, 'tracker': 'botsort.yaml', 'val_interval': 20, 
    'text_model': 'mobileclip:blt', 'load_vp': False, 'train_pe_path': 'None'
}
for k, v in DEFAULT_CFG_DICT.items():
    if isinstance(v, str) and v.lower() == "none":
        DEFAULT_CFG_DICT[k] = None
DEFAULT_CFG_KEYS = DEFAULT_CFG_DICT.keys()
DEFAULT_CFG = IterableSimpleNamespace(**DEFAULT_CFG_DICT)


class Retry(contextlib.ContextDecorator):
    """
    Retry class for function execution with exponential backoff.

    Can be used as a decorator to retry a function on exceptions, up to a specified number of times with an
    exponentially increasing delay between retries.

    Examples:
        Example usage as a decorator:
        >>> @Retry(times=3, delay=2)
        >>> def test_func():
        >>> # Replace with function logic that may raise exceptions
        >>>     return True
    """

    def __init__(self, times=3, delay=2):
        """Initialize Retry class with specified number of retries and delay."""
        self.times = times
        self.delay = delay
        self._attempts = 0

    def __call__(self, func):
        """Decorator implementation for Retry with exponential backoff."""

        def wrapped_func(*args, **kwargs):
            """Applies retries to the decorated function or method."""
            self._attempts = 0
            while self._attempts < self.times:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    self._attempts += 1
                    print(f"Retry {self._attempts}/{self.times} failed: {e}")
                    if self._attempts >= self.times:
                        raise e
                    time.sleep(self.delay * (2**self._attempts))  # exponential backoff delay

        return wrapped_func


def threaded(func):
    """
    Multi-threads a target function by default and returns the thread or function result.

    Use as @threaded decorator. The function runs in a separate thread unless 'threaded=False' is passed.
    """

    def wrapper(*args, **kwargs):
        """Multi-threads a given function based on 'threaded' kwarg and returns the thread or function result."""
        if kwargs.pop("threaded", True):  # run in thread
            thread = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
            thread.start()
            return thread
        else:
            return func(*args, **kwargs)

    return wrapper


def clean_url(url):
    """Strip auth from URL, i.e. https://url.com/file.txt?auth -> https://url.com/file.txt."""
    url = Path(url).as_posix().replace(":/", "://")  # Pathlib turns :// -> :/, as_posix() for Windows
    return urllib.parse.unquote(url).split("?")[0]  # '%2F' to '/', split https://url.com/file.txt?auth


def url2file(url):
    """Convert URL to filename, i.e. https://url.com/file.txt?auth -> file.txt."""
    return Path(clean_url(url)).name


# Run below code on utils init ------------------------------------------------------------------------------------

# Apply monkey patches
from visaionlibrary.utils.yoloe.utils.patches import imread, imshow, imwrite, torch_load, torch_save

torch.load = torch_load
torch.save = torch_save
if WINDOWS:
    # Apply cv2 patches for non-ASCII and non-UTF characters in image paths
    cv2.imread, cv2.imwrite, cv2.imshow = imread, imwrite, imshow



from .pipeline_switch_hook import VisaionPipelineSwitchHook
from .yolov5_param_scheduler_hook import YOLOv5IterParamSchedulerHook
__all__ = ['VisaionPipelineSwitchHook', 'YOLOv5IterParamSchedulerHook']


from visaionlibrary.datasets.transforms.loading import LoadSegAnnotations
from visaionlibrary.datasets.transforms.process import RandomRescale
from visaionlibrary.datasets.transforms.region_sampler import SegRegionSampler
from visaionlibrary.datasets.transforms.formatting import PackSegInputsVisaion

__all__ = ["LoadSegAnnotations", "RandomRescale", "SegRegionSampler", "PackSegInputsVisaion"]


from dataclasses import dataclass

import numpy as np
from PIL import Image

@dataclass
class SourceTypes:
    """
    Class to represent various types of input sources for predictions.

    This class uses dataclass to define boolean flags for different types of input sources that can be used for
    making predictions with YOLO models.

    Attributes:
        stream (bool): Flag indicating if the input source is a video stream.
        screenshot (bool): Flag indicating if the input source is a screenshot.
        from_img (bool): Flag indicating if the input source is an image file.

    Examples:
        >>> source_types = SourceTypes(stream=True, screenshot=False, from_img=False)
        >>> print(source_types.stream)
        True
        >>> print(source_types.from_img)
        False
    """

    stream: bool = False
    screenshot: bool = False
    from_img: bool = False
    tensor: bool = False


class LoadPilAndNumpy:
    """
    Load images from PIL and Numpy arrays for batch processing.

    This class manages loading and pre-processing of image data from both PIL and Numpy formats. It performs basic
    validation and format conversion to ensure that the images are in the required format for downstream processing.

    Attributes:
        paths (List[str]): List of image paths or autogenerated filenames.
        im0 (List[np.ndarray]): List of images stored as Numpy arrays.
        mode (str): Type of data being processed, set to 'image'.
        bs (int): Batch size, equivalent to the length of `im0`.

    Methods:
        _single_check: Validate and format a single image to a Numpy array.

    Examples:
        >>> from PIL import Image
        >>> import numpy as np
        >>> pil_img = Image.new("RGB", (100, 100))
        >>> np_img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
        >>> loader = LoadPilAndNumpy([pil_img, np_img])
        >>> paths, images, _ = next(iter(loader))
        >>> print(f"Loaded {len(images)} images")
        Loaded 2 images
    """

    def __init__(self, im0):
        """Initializes a loader for PIL and Numpy images, converting inputs to a standardized format."""
        if not isinstance(im0, list):
            im0 = [im0]
        # use `image{i}.jpg` when Image.filename returns an empty path.
        self.paths = [getattr(im, "filename", "") or f"image{i}.jpg" for i, im in enumerate(im0)]
        self.im0 = [self._single_check(im) for im in im0]
        self.mode = "image"
        self.bs = len(self.im0)

    @staticmethod
    def _single_check(im):
        """Validate and format an image to numpy array, ensuring RGB order and contiguous memory."""
        assert isinstance(im, (Image.Image, np.ndarray)), f"Expected PIL/np.ndarray image type, but got {type(im)}"
        if isinstance(im, Image.Image):
            if im.mode != "RGB":
                im = im.convert("RGB")
            im = np.asarray(im)[:, :, ::-1]
            im = np.ascontiguousarray(im)  # contiguous
        return im

    def __len__(self):
        """Returns the length of the 'im0' attribute, representing the number of loaded images."""
        return len(self.im0)

    def __next__(self):
        """Returns the next batch of images, paths, and metadata for processing."""
        if self.count == 1:  # loop only once as it's batch inference
            raise StopIteration
        self.count += 1
        return self.paths, self.im0, [""] * self.bs

    def __iter__(self):
        """Iterates through PIL/numpy images, yielding paths, raw images, and metadata for processing."""
        self.count = 0
        return self



"""
RTMDetInsSepBNHead for Visaion
在rtmdet_s_ins模板中使用了RTMDetInsSepBNHead, 在使用过程中发现, 其父类RTMDetInsHead中_bbox_mask_post_process方法的中的
if rescale:
    ori_h, ori_w = img_meta['ori_shape'][:2]
    mask_logits = F.interpolate(
        mask_logits,
        size=[
            math.ceil(mask_logits.shape[-2] * scale_factor[0]),
            math.ceil(mask_logits.shape[-1] * scale_factor[1])
        ],
        mode='bilinear',
        align_corners=False)[..., :ori_h, :ori_w]

有错误, 需要修改
"""
from typing import Optional
import math

import torch
import torch.nn.functional as F

from mmcv.ops import batched_nms
from mmengine.structures import InstanceData
from mmengine.registry import MODELS
from mmdet.utils import ConfigType
from mmdet.structures.bbox import (get_box_tensor, get_box_wh, scale_boxes)
from mmdet.models.dense_heads.rtmdet_ins_head import RTMDetInsSepBNHead


@MODELS.register_module()
class VisaionRTMDetInsSepBNHead(RTMDetInsSepBNHead):
    def _bbox_mask_post_process(
            self,
            results: InstanceData,
            mask_feat,
            cfg: ConfigType,
            rescale: bool = False,
            with_nms: bool = True,
            img_meta: Optional[dict] = None) -> InstanceData:
        """bbox and mask post-processing method.

        The boxes would be rescaled to the original image scale and do
        the nms operation. Usually `with_nms` is False is used for aug test.

        Args:
            results (:obj:`InstaceData`): Detection instance results,
                each item has shape (num_bboxes, ).
            cfg (ConfigDict): Test / postprocessing configuration,
                if None, test_cfg would be used.
            rescale (bool): If True, return boxes in original image space.
                Default to False.
            with_nms (bool): If True, do nms before return boxes.
                Default to True.
            img_meta (dict, optional): Image meta info. Defaults to None.

        Returns:
            :obj:`InstanceData`: Detection results of each image
            after the post process.
            Each item usually contains following keys.

                - scores (Tensor): Classification scores, has a shape
                  (num_instance, )
                - labels (Tensor): Labels of bboxes, has a shape
                  (num_instances, ).
                - bboxes (Tensor): Has a shape (num_instances, 4),
                  the last dimension 4 arrange as (x1, y1, x2, y2).
                - masks (Tensor): Has a shape (num_instances, h, w).
        """
        stride = self.prior_generator.strides[0][0]
        if rescale:
            assert img_meta.get('scale_factor') is not None
            scale_factor = [1 / s for s in img_meta['scale_factor']]
            results.bboxes = scale_boxes(results.bboxes, scale_factor)

        if hasattr(results, 'score_factors'):
            # TODO： Add sqrt operation in order to be consistent with
            #  the paper.
            score_factors = results.pop('score_factors')
            results.scores = results.scores * score_factors

        # filter small size bboxes
        if cfg.get('min_bbox_size', -1) >= 0:
            w, h = get_box_wh(results.bboxes)
            valid_mask = (w > cfg.min_bbox_size) & (h > cfg.min_bbox_size)
            if not valid_mask.all():
                results = results[valid_mask]

        # TODO: deal with `with_nms` and `nms_cfg=None` in test_cfg
        assert with_nms, 'with_nms must be True for RTMDet-Ins'
        if results.bboxes.numel() > 0:
            bboxes = get_box_tensor(results.bboxes)
            det_bboxes, keep_idxs = batched_nms(bboxes, results.scores,
                                                results.labels, cfg.nms)
            results = results[keep_idxs]
            # some nms would reweight the score, such as softnms
            results.scores = det_bboxes[:, -1]
            results = results[:cfg.max_per_img]

            # process masks
            mask_logits = self._mask_predict_by_feat_single(
                mask_feat, results.kernels, results.priors)

            mask_logits = F.interpolate(
                mask_logits.unsqueeze(0), scale_factor=stride, mode='bilinear')
            if rescale:
                ori_h, ori_w = img_meta['ori_shape'][:2]
                mask_logits = F.interpolate(
                    mask_logits,
                    size=[
                        math.ceil(mask_logits.shape[-2] * scale_factor[1]),
                        math.ceil(mask_logits.shape[-1] * scale_factor[0])
                    ],
                    mode='bilinear',
                    align_corners=False)[..., :ori_h, :ori_w]
            masks = mask_logits.sigmoid().squeeze(0)
            masks = masks > cfg.mask_thr_binary
            results.masks = masks
        else:
            h, w = img_meta['ori_shape'][:2] if rescale else img_meta[
                'img_shape'][:2]
            results.masks = torch.zeros(
                size=(results.bboxes.shape[0], h, w),
                dtype=torch.bool,
                device=results.bboxes.device)

        return results


from typing import List
import os
import os.path as osp
import json
import cv2
import time

from mmengine.registry import DATASETS
from mmyolo.datasets import YOLOv5CocoDataset
from mmengine.logging import print_log
from mmengine import load

from visaionlibrary.datasets.utils import SampleStatus

@DATASETS.register_module()
class VisaionDetDataset(YOLOv5CocoDataset):
    def load_data_list(self) -> List[dict]:
        """
        将visaion格式的数据集转换为coco格式
        """
        json_file = osp.join(f"visaion_det_{time.time()}.json")

        if isinstance(self.data_root, str) and "::" not in self.data_root:
            self.data_root = [self.data_root]
        elif isinstance(self.data_root, str) and "::" in self.data_root:
            self.data_root = self.data_root.split("::") 
        elif isinstance(self.data_root, list):
            pass
        else:
            raise ValueError(f"data_root must be a string of dataset path or a list of dataset path, but got {self.data_root}")

        # dataset可以是文件夹和json文件
        # - 文件夹是visaion后端格式
        # - json是visaion前端兼容格式
        self.ann_file = json_file
        print_log(f"convert visaion format dataset to coco format ...", "current")
        if self.data_root[0].endswith("json"):
            self.convert_visaion_json_to_coco_det(self.ann_file)
        else:
            self.convert_visaion_folder_to_coco_det(self.ann_file)

        # 调用父类的load_data_list方法
        data_list = super().load_data_list()

        # 判断数据集中每个数据是前景还是背景
        print_log(f"collect samples type ...", "current")
        self.samples_type = []
        for data in data_list:
            if len(data['instances']) == 0:
                self.samples_type.append(SampleStatus.BACKGROUND)
            else:
                self.samples_type.append(SampleStatus.FOREGROUND)
        
        os.remove(json_file)
        return data_list

    
    def convert_visaion_folder_to_coco_det(self, output_file: str):
        """
        将visaion后端格式转换为coco格式
        Args:
            data_roots: 数据集路径
            output_file: 转换成coco数据集的json文件
        """
        coco_data = {
            "info": {
                "description": "Converted from visaion format",
                "version": "1.0",
                "year": 2025,
                "contributor": "visaion2coco",
                "date_created": "2025/05/11"
            },
            "licenses": [{
                "url": "http://creativecommons.org/licenses/by-nc-sa/2.0/",
                "id": 1,
                "name": "Attribution-NonCommercial-ShareAlike License"
            }],
            "images": [],
            "annotations": [],
            "categories": []
        }

        # 创建类别映射
        category_name_to_id = {}
        category_id = 0  # 从0开始
        annotation_id = 0  # 从0开始

        # 从数据集中获取类别信息, 原来是转coco获取的, 那样会有bug
        for class_name in self.metainfo['classes']:
            if class_name not in category_name_to_id:
                category_name_to_id[class_name] = category_id
                coco_data['categories'].append({
                    "id": category_id,
                    "name": class_name,
                    "supercategory": "none"
                })
                category_id += 1

        for data_root in self.data_root:
            for file_name in os.listdir(data_root):
                if not file_name.endswith(('.png', '.bmp', '.jpg', '.jpeg')):
                    continue

                img_path = os.path.join(data_root, file_name)
                img = cv2.imread(img_path)
                if img is None:
                    continue
                
                height, width = img.shape[:2]
                
                # 添加图片信息
                image_id = len(coco_data['images'])  # 从0开始
                coco_data['images'].append({
                    "id": image_id,
                    "file_name": os.path.join(data_root, file_name),
                    "width": width,
                    "height": height,
                    "date_captured": "2024-05-03"
                })
                
                # 检查是否有对应的json文件
                json_file = file_name.rsplit('.', 1)[0] + '.json'
                json_path = os.path.join(data_root, json_file)
                if not os.path.exists(json_path):
                    json_file = file_name.rsplit('.', 1)[0] + '.mix.json'
                    json_path = os.path.join(data_root, json_file)
                
                if os.path.exists(json_path):
                    # 读取json文件获取标注信息
                    with open(json_path, 'r') as f:
                        ann_data = json.load(f)
                    
                    # 添加标注信息
                    for detection in ann_data.get('detections', []):
                        bbox = detection['boundingBox']  # [x, y, w, h]
                        label_name = detection['labelName']
                        
                        if label_name in category_name_to_id:
                            coco_data['annotations'].append({
                                "id": annotation_id,
                                "image_id": image_id,
                                "category_id": category_name_to_id[label_name],
                                "bbox": bbox,
                                "area": bbox[2] * bbox[3],
                                "segmentation": [],
                                "iscrowd": 0
                            })
                            annotation_id += 1

        # 保存为coco格式的json文件
        with open(output_file, 'w') as f:
            json.dump(coco_data, f, indent=2)

    def convert_visaion_json_to_coco_det(self, output_file: str):
        """
        将json文件转换为coco格式
        Args:
            input_file: 输入的json文件路径
            output_file: 输出的coco格式json文件路径
        
        self.data_root: List[str]  数据集路径, eg. ['/visaion/projects/1/datasets/5617.json', ...]
            1116.json内容如下:
            {
                'meta_info': {
                    'url_prefix': 'projects/1/samples'
                },
                'dataset': [
                    {
                        'height': 429, 
                        'width': 473, 
                        'sample_id': 2882, 
                        'mem_size': 1623336, 
                        'mix_annotation_path': '0000001_CCD31-0_20220604172015_8bb121aa92.json',
                        'data': [
                            {
                                'img_path': '0000001_CCD31-0_20220604172015_a1076a68f4.bmp',
                            }
                        ]
                    },
                    ...,
                ]
            }
        
        """
        VISAION_DIR = os.environ['VISAION_DIR']
        assert VISAION_DIR is not None, "VISAION_DIR is not set"

        coco_data = {
            "info": {
                "description": "Converted from visaion format",
                "version": "1.0",
                "year": 2025,
                "contributor": "visaion2coco",
                "date_created": "2025/05/11"
            },
            "licenses": [{
                "url": "http://creativecommons.org/licenses/by-nc-sa/2.0/",
                "id": 1,
                "name": "Attribution-NonCommercial-ShareAlike License"
            }],
            "images": [],
            "annotations": [],
            "categories": []
        }

        # 创建类别映射
        category_name_to_id = {}
        category_id = 0  # 从0开始
        annotation_id = 0  # 从0开始

        # 从数据集中获取类别信息, 原来是转coco获取的, 那样会有bug
        for class_name in self.metainfo['classes']:
            if class_name not in category_name_to_id:
                category_name_to_id[class_name] = category_id
                coco_data['categories'].append({
                    "id": category_id,
                    "name": class_name,
                    "supercategory": "none"
                })
                category_id += 1

        for dataset_json_path in self.data_root:
            # 读取数据集内容, 即一个json文件中的内容
            dataset_json_dict = load(dataset_json_path)
            all_sample = dataset_json_dict['dataset']
            url_prefix = dataset_json_dict['meta_info']['url_prefix']

            for sample in all_sample:
                # 添加图片信息
                image_id = len(coco_data['images'])
                img_path = os.path.join(VISAION_DIR, url_prefix, sample['data'][0]['img_path'])
                
                coco_data['images'].append({
                    "id": image_id,
                    "file_name": img_path,
                    "width": sample['width'],
                    "height": sample['height'],
                    "date_captured": "2024-05-03"
                })

                # 读取标注文件
                # 如果sample['mix_annotation_path'] is None or sample['mix_annotation_path'] == "", 则跳过
                if sample['mix_annotation_path'] is None or sample['mix_annotation_path'] == "":
                    continue
                ann_path = os.path.join(VISAION_DIR, url_prefix, sample['mix_annotation_path'])
                if os.path.exists(ann_path):
                    with open(ann_path, 'r') as f:
                        ann_data = json.load(f)
                    
                    # 添加标注信息
                    for detection in ann_data.get('detections', []):
                        bbox = detection['boundingBox']  # [x, y, w, h]
                        label_name = detection['labelName']
                        
                        if label_name in category_name_to_id:
                            coco_data['annotations'].append({
                                "id": annotation_id,
                                "image_id": image_id,
                                "category_id": category_name_to_id[label_name],
                                "bbox": bbox,
                                "area": bbox[2] * bbox[3],
                                "segmentation": [],
                                "iscrowd": 0
                            })
                            annotation_id += 1

        # 保存为coco格式的json文件
        with open(output_file, 'w') as f:
            json.dump(coco_data, f, indent=2)